# sage_setup: distribution = sagemath-repl
"""
Marimo backend for the Sage Rich Output System

This module defines the Marimo backend for
:mod:`sage.repl.rich_output`.
"""

# ****************************************************************************
#       Copyright (C) 2026 Shahzeb Imran <s.imran@tuta.io>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import base64
import html
from sage.repl.rich_output import get_display_manager  # type: ignore
from sage.repl.rich_output.backend_base import BackendBase
from sage.repl.rich_output.output_basic import OutputBase
from sage.repl.rich_output.output_catalog import (
    OutputAsciiArt,
    OutputHtml,
    OutputImageGif,
    OutputImageJpg,
    OutputImagePng,
    OutputImageSvg,
    OutputLatex,
    OutputPlainText,
    OutputSceneThreejs,
    OutputUnicodeArt,
)
from sage.repl.rich_output.preferences import DisplayPreferences
from typing import Any

# ensures only the following things are part of the public interface
__all__ = ["check_and_load_marimo", "BackendMarimo"]


def check_and_load_marimo():
    """Switches to BackendMarimo if running in an interactive marimo notebook."""
    try:
        import marimo

        if marimo.running_in_notebook():
            print("Interactive marimo session detected. Loading sage marimo backend.")
            from sage.repl.rich_output.backend_marimo import BackendMarimo

            get_display_manager().switch_backend(BackendMarimo())
    except ImportError:
        pass


class BackendMarimo(BackendBase):
    """
    Marimo backend

    EXAMPLES::

        # in a marimo notebook

        from sage.repl.rich_output.backend_marimo import BackendMarimo
        from sage.repl.rich_output import get_display_manager
        dm = get_display_manager()
        dm.switch_backend(BackendMarimo())
        print(dm.get_instance())

        # output:
        The Sage display manager using the Marimo Notebook backend

        print(dm.preferences())

        # output:
        Display preferences:
        * align_latex = center
        * graphics is not specified
        * supplemental_plot is not specified
        * text = latex
    """

    _SUPPORTED_OUTPUT_CLASSES = (
        OutputPlainText,
        OutputAsciiArt,
        OutputUnicodeArt,
        OutputHtml,
        OutputImagePng,
        OutputImageJpg,
        OutputImageGif,
        OutputImageSvg,
        OutputSceneThreejs,
        OutputLatex,
    )

    def _repr_(self) -> str:
        return "Marimo Notebook"

    def install(self) -> None:
        from marimo._output import formatting
        from sage.structure.sage_object import SageObject

        # register formatters with marimo
        @formatting.formatter(SageObject)
        def _show_sage_object(sobj: SageObject) -> tuple[str, str]:
            """
            handles cases where we have a SageObject and we have to get the rich_output
            before rendering it
            """
            plain_text, rich_output = get_display_manager()._rich_output_formatter(
                sobj, dict()
            )
            return self._render_rich_output(rich_output, fallback_obj=sobj)

        def _show_output_base(obj: OutputBase) -> tuple[str, str]:
            """
            handles cases where we already have the rich_output
            e.g. when .show() is invoked and we register the output through
            a call to display_immediately
            """
            return self._render_rich_output(obj)

        for cls in self._SUPPORTED_OUTPUT_CLASSES:
            formatting.formatter(cls)(_show_output_base)

    def default_preferences(self) -> DisplayPreferences:
        return DisplayPreferences(text="latex", align_latex="center")

    def supported_output(self) -> set:
        return set(self._SUPPORTED_OUTPUT_CLASSES)

    def _format_html_image(self, image_format: str, image: bytes) -> tuple[str, str]:
        b64 = base64.b64encode(image).decode("ascii")

        img_html = f"""
       <div style="resize: vertical; overflow: hidden; width: 100%; display: flex; align-items: center; justify-content: center;">
         <img src="data:image/{image_format}; base64,{b64}"
           style="max-width: 100%; max-height: 100%; object-fit: contain;" />
       </div>
       """

        return ("text/html", img_html)

    def _render_rich_output(
        self, rich_output: OutputBase, fallback_obj: Any = None
    ) -> tuple[str, str]:
        import marimo as mo

        match rich_output:
            case OutputHtml():
                html_str = rich_output.html.get_str()
                # the default sage latex formatter outputs latex with <html> tags
                # we strip them if present
                if html_str.startswith("<html>") and html_str.endswith("</html>"):
                    html_str = html_str[len("<html>") : -len("</html>")]
                return (
                    "text/html",
                    mo.md(f"""{html_str}""").text,
                )
            # we wrap graphics outputs in a div to allow resizing
            case OutputSceneThreejs():
                escaped_html = html.escape(rich_output.html.get_str())

                iframe = f"""
                <div style="resize: vertical; overflow: hidden; width: 100%; min-width: 100%; max-width: 100%; position: relative; min-height: 150px;">
                    <iframe srcdoc="{escaped_html}"
                            style="position: absolute; top: 0; left: 0; width: 100%; height: 100%; border: 0;">
                    </iframe>
                </div>
                """
                return ("text/html", iframe)

            case OutputImagePng():
                return self._format_html_image("png", rich_output.png.get())

            case OutputImageJpg():
                return self._format_html_image("jpg", rich_output.jpg.get())

            case OutputImageGif():
                return self._format_html_image("gif", rich_output.gif.get())

            case OutputImageSvg():
                svg_str = rich_output.svg.get_str()

                svg_html = f"""
                <div class="resizable-svg-wrapper" style="resize: vertical; overflow: hidden; width: 100%; display: flex; align-items: center; justify-content: center;">
                    <style>
                        .resizable-svg-wrapper > svg {{
                            max-width: 100% !important;
                            max-height: 100% !important;
                            height: auto !important;
                            object-fit: contain !important;
                        }}
                    </style>
                    {svg_str}
                </div>
                """
                return ("text/html", svg_html)

            # since we have OutputHtml enabled, the sage output formatter automatically
            # outputs rich output as MathJax HTML.
            # This one is for raw latex output
            case OutputLatex():
                return (
                    "text/html",
                    mo.md(f"""\\[{rich_output.latex.get_str()}\\]""").text,
                )

            case OutputAsciiArt():
                return (
                    "text/html",
                    f"<pre>{html.escape(rich_output.ascii_art.get_str())}</pre>",
                )

            case OutputUnicodeArt():
                return (
                    "text/html",
                    f"<pre>{html.escape(rich_output.unicode_art.get_str())}</pre>",
                )

            case OutputPlainText():
                return ("text/plain", rich_output.text.get_str())

        # fallback: try latex output first and if that fails, plaintext
        if fallback_obj is not None:
            try:
                return (
                    "text/html",
                    mo.md(f"""\\[{fallback_obj._latex_()}\\]""").text,
                )
            except (AttributeError, TypeError):
                return ("text/plain", repr(fallback_obj))

        return ("text/plain", repr(rich_output))

    def display_immediately(
        self, plain_text: OutputPlainText, rich_output: OutputBase
    ) -> None:
        import marimo as mo

        mo.output.append(rich_output)

    def threejs_offline_scripts(self) -> str:
        return get_display_manager().threejs_scripts(online=True)



# sage_setup: distribution = sagemath-repl
# sage.doctest: needs sage.combinat sage.graphs sage.groups
r"""
This file contains doctests for the Chapter "k-Schur function primer"
for the book "k-Schur functions and affine Schubert calculus"
by Thomas Lam, Luc Lapointe, Jennifer Morse, Anne Schilling, Mark Shimozono,
and Mike Zabrocki, :arxiv:`1301.3569`.
The code was written by Anne Schilling and Mike Zabrocki, 2012 and 2013.

IF IT BECOMES NECESSARY TO CHANGE ANY TESTS IN THIS FILE, THERE
NEEDS TO BE A ONE-YEAR DEPRECATION PERIOD. ALSO, PLEASE IN THIS CASE
CONTACT Anne Schilling (anne@math.ucdavis.edu) AND Mike Zabrocki
(zabrocki@mathstat.yorku.ca) REGARDING THE CHANGES!
"""
"""
Sage example in ./kschurnotes/notes-mike-anne.tex, line 198::

    sage: P = Partitions(4); P
    Partitions of the integer 4
    sage: P.list()
    [[4], [3, 1], [2, 2], [2, 1, 1], [1, 1, 1, 1]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 205::

    sage: [p for p in P]
    [[4], [3, 1], [2, 2], [2, 1, 1], [1, 1, 1, 1]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 210::

    sage: la=Partition([2,2]); mu=Partition([3,1])
    sage: mu.dominates(la)
    True

Sage example in ./kschurnotes/notes-mike-anne.tex, line 216::

    sage: ord = lambda x,y: y.dominates(x)
    sage: P = Poset([Partitions(6), ord], facade=True)
    sage: H = P.hasse_diagram()

Sage example in ./kschurnotes/notes-mike-anne.tex, line 228::

    sage: la=Partition([4,3,3,3,2,2,1])
    sage: la.conjugate()
    [7, 6, 4, 1]
    sage: la.k_split(4)
    [[4], [3, 3], [3, 2], [2, 1]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 236::

    sage: p = SkewPartition([[2,1],[1]])
    sage: p.is_connected()
    False

Sage example in ./kschurnotes/notes-mike-anne.tex, line 334::

    sage: la = Partition([4,3,3,3,2,2,1])
    sage: kappa = la.k_skew(4); kappa
    [12, 8, 5, 5, 2, 2, 1] / [8, 5, 2, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 340::

    sage: kappa.row_lengths()
    [4, 3, 3, 3, 2, 2, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 345::

    sage: tau = Core([12,8,5,5,2,2,1],5)
    sage: mu = tau.to_bounded_partition(); mu
    [4, 3, 3, 3, 2, 2, 1]
    sage: mu.to_core(4)
    [12, 8, 5, 5, 2, 2, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 353::

    sage: Cores(3,6).list()
    [[6, 4, 2], [5, 3, 1, 1], [4, 2, 2, 1, 1], [3, 3, 2, 2, 1, 1]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 398::

    sage: W = WeylGroup(['A',4,1])      # long time (5.47 s, 2013)
    sage: S = W.simple_reflections()    # long time
    sage: [s.reduced_word() for s in S] # long time
    [[0], [1], [2], [3], [4]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 406::

    sage: # long time
    sage: w = W.an_element(); w
    [ 2  0  0  1 -2]
    [ 2  0  0  0 -1]
    [ 1  1  0  0 -1]
    [ 1  0  1  0 -1]
    [ 1  0  0  1 -1]
    sage: w.reduced_word()
    [0, 1, 2, 3, 4]
    sage: w = W.from_reduced_word([2,1,0])
    sage: w.is_affine_grassmannian()
    True

Sage example in ./kschurnotes/notes-mike-anne.tex, line 464::

    sage: c = Core([7,3,1],5)
    sage: c.affine_symmetric_group_simple_action(2)
    [8, 4, 1, 1]
    sage: c.affine_symmetric_group_simple_action(0)
    [7, 3, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 474::

    sage: k=4; length=3
    sage: W = WeylGroup(['A',k,1])
    sage: G = W.affine_grassmannian_elements_of_given_length(length)
    sage: [w.reduced_word() for w in G]
    [[2, 1, 0], [4, 1, 0], [3, 4, 0]]

    sage: C = Cores(k+1,length)
    sage: [c.to_grassmannian().reduced_word() for c in C]
    [[2, 1, 0], [4, 1, 0], [3, 4, 0]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 543::

    sage: la = Partition([4,3,3,3,2,2,1])
    sage: c = la.to_core(4); c
    [12, 8, 5, 5, 2, 2, 1]
    sage: W = WeylGroup(['A',4,1])
    sage: w = W.from_reduced_word([4,1,0,2,1,4,3,2,0,4,3,1,0,4,3,2,1,0])
    sage: c.to_grassmannian() == w
    True

Sage example in ./kschurnotes/notes-mike-anne.tex, line 643::

sage: A = AffinePermutationGroup(['A',2,1])
sage: w = A([-2,0,8])
sage: w.reduced_word()
[1, 0, 2, 1, 0]
sage: w.to_core()
[5, 3, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 767::

    sage: la = Partition([4,3,3,3,2,2,1])
    sage: la.k_conjugate(4)
    [3, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1065::

    sage: c = Core([3,1,1],3)
    sage: c.weak_covers()
    [[4, 2, 1, 1]]
    sage: c.strong_covers()
    [[5, 3, 1], [4, 2, 1, 1], [3, 2, 2, 1, 1]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1073::

    sage: kappa = Core([4,1],4)
    sage: tau = Core([2,1],4)
    sage: tau.weak_le(kappa)
    False
    sage: tau.strong_le(kappa)
    True

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1082::

    sage: C = sum(([c for c in Cores(4,m)] for m in range(7)),[])
    sage: ord = lambda x,y: x.weak_le(y)
    sage: P = Poset([C, ord], cover_relations = False)  # long time (3.99 s, 2013)
    sage: H = P.hasse_diagram()                         # long time
    sage: view(H)  # not tested

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1253::

    sage: Sym = SymmetricFunctions(QQ)
    sage: h = Sym.homogeneous()
    sage: m = Sym.monomial()

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1259::

    sage: f = h[3,1]+h[2,2]
    sage: m(f)
    10*m[1, 1, 1, 1] + 7*m[2, 1, 1] + 5*m[2, 2] + 4*m[3, 1] + 2*m[4]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1266::

    sage: f.scalar(h[2,1,1])
    7
    sage: m(f).coefficient([2,1,1])
    7

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1274::

    sage: p = Sym.power()
    sage: e = Sym.elementary()
    sage: sum( (-1)**(i-1)*e[4-i]*p[i] for i in range(1,4) ) - p[4]
    4*e[4]
    sage: sum( (-1)**(i-1)*p[i]*e[4-i] for i in range(1,4) ) - p[4]
    1/6*p[1, 1, 1, 1] - p[2, 1, 1] + 1/2*p[2, 2] + 4/3*p[3, 1] - p[4]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1327::

    sage: Sym = SymmetricFunctions(QQ)
    sage: s = Sym.schur()
    sage: m = Sym.monomial()
    sage: h = Sym.homogeneous()
    sage: m(s[1,1,1])
    m[1, 1, 1]
    sage: h(s[1,1,1])
    h[1, 1, 1] - 2*h[2, 1] + h[3]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1338::

    sage: p = Sym.power()
    sage: s = Sym.schur()
    sage: p(s[1,1,1])
    1/6*p[1, 1, 1] - 1/2*p[2, 1] + 1/3*p[3]
    sage: p(s[2,1])
    1/3*p[1, 1, 1] - 1/3*p[3]
    sage: p(s[3])
    1/6*p[1, 1, 1] + 1/2*p[2, 1] + 1/3*p[3]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1349::

    sage: s[2,1].scalar(s[1,1,1])
    0
    sage: s[2,1].scalar(s[2,1])
    1

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1517::

  sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
  sage: Qp = Sym.hall_littlewood().Qp()
  sage: Qp.base_ring()
  Fraction Field of Univariate Polynomial Ring in t over Rational Field
  sage: s = Sym.schur()
  sage: s(Qp[1,1,1])
  s[1, 1, 1] + (t^2+t)*s[2, 1] + t^3*s[3]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1530::

  sage: t = Qp.t
  sage: s[2,1].scalar(s[3].theta_qt(t,0))
  t^2 - t

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1536::

  sage: s(Qp([1,1])).hl_creation_operator([3])
  s[3, 1, 1] + t*s[3, 2] + (t^2+t)*s[4, 1] + t^3*s[5]
  sage: s(Qp([3,1,1]))
  s[3, 1, 1] + t*s[3, 2] + (t^2+t)*s[4, 1] + t^3*s[5]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1568::

  sage: Sym = SymmetricFunctions(FractionField(QQ['q,t']))
  sage: Mac = Sym.macdonald()
  sage: H = Mac.H()
  sage: s = Sym.schur()
  sage: for la in Partitions(3):
  ....:     print("H {} = {}".format(la, s(H(la))))
  H [3] = q^3*s[1, 1, 1] + (q^2+q)*s[2, 1] + s[3]
  H [2, 1] = q*s[1, 1, 1] + (q*t+1)*s[2, 1] + t*s[3]
  H [1, 1, 1] = s[1, 1, 1] + (t^2+t)*s[2, 1] + t^3*s[3]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1581::

  sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
  sage: Mac = Sym.macdonald(q=0)
  sage: H = Mac.H()
  sage: s = Sym.schur()
  sage: for la in Partitions(3):
  ....:     print("H {} = {}".format(la, s(H(la))))
  H [3] = s[3]
  H [2, 1] = s[2, 1] + t*s[3]
  H [1, 1, 1] = s[1, 1, 1] + (t^2+t)*s[2, 1] + t^3*s[3]
  sage: Qp = Sym.hall_littlewood().Qp()
  sage: s(Qp[1, 1, 1])
  s[1, 1, 1] + (t^2+t)*s[2, 1] + t^3*s[3]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1596::

  sage: Sym = SymmetricFunctions(FractionField(QQ['q']))
  sage: Mac = Sym.macdonald(t=0)
  sage: H = Mac.H()
  sage: s = Sym.schur()
  sage: for la in Partitions(3):
  ....:     print("H {} = {}".format(la, s(H(la))))
  H [3] = q^3*s[1, 1, 1] + (q^2+q)*s[2, 1] + s[3]
  H [2, 1] = q*s[1, 1, 1] + s[2, 1]
  H [1, 1, 1] = s[1, 1, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1717::

    sage: t = Tableau([[1,1,1,2,3,7],[2,2,3,5],[3,4],[4,5],[6]])
    sage: t.charge()
    9

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1881::

    sage: la = Partition([2,2])
    sage: la.k_conjugate(2).conjugate()
    [4]
    sage: la = Partition([2,1,1])
    sage: la.k_conjugate(2).conjugate()
    [3, 1]
    sage: la = Partition([1,1,1,1])
    sage: la.k_conjugate(2).conjugate()
    [2, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1893::

    sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
    sage: ks = Sym.kschur(2)
    sage: ks[2,2].omega_t_inverse()
    1/t^2*ks2[1, 1, 1, 1]
    sage: ks[2,1,1].omega_t_inverse()
    1/t*ks2[2, 1, 1]
    sage: ks[1,1,1,1].omega_t_inverse()
    1/t^2*ks2[2, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 1904::

    sage: Sym = SymmetricFunctions(FractionField(QQ['q,t']))
    sage: H = Sym.macdonald().H()
    sage: ks = Sym.kschur(2)
    sage: ks(H[2,2])
    q^2*ks2[1, 1, 1, 1] + (q*t+q)*ks2[2, 1, 1] + ks2[2, 2]
    sage: ks(H[2,1,1])
    q*ks2[1, 1, 1, 1] + (q*t^2+1)*ks2[2, 1, 1] + t*ks2[2, 2]
    sage: ks(H[1,1,1,1])
    ks2[1, 1, 1, 1] + (t^3+t^2)*ks2[2, 1, 1] + t^4*ks2[2, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2174::

    sage: SemistandardTableaux([5,2],[4,2,1]).list()
    [[[1, 1, 1, 1, 2], [2, 3]], [[1, 1, 1, 1, 3], [2, 2]]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2179::

    sage: P = Partitions(4)
    sage: P.list()
    [[4], [3, 1], [2, 2], [2, 1, 1], [1, 1, 1, 1]]
    sage: n = P.cardinality(); n
    5
    sage: K = matrix(QQ,n,n,
    ....:           [[SemistandardTableaux(la,mu).cardinality()
    ....:            for mu in P] for la in P])
    sage: K
    [1 1 1 1 1]
    [0 1 1 2 3]
    [0 0 1 1 2]
    [0 0 0 1 3]
    [0 0 0 0 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2376::

    sage: T = WeakTableaux(6, [5,3], [4,3,1])
    sage: T.list()
    [[[1, 1, 1, 1, 3], [2, 2, 2]], [[1, 1, 1, 1, 2], [2, 2, 3]]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2382::

    sage: k = 3
    sage: c = Core([5,2,1], k+1)
    sage: la = c.to_bounded_partition(); la
    [3, 2, 1]
    sage: for mu in Partitions(la.size(), max_part = 3):
    ....:     T = WeakTableaux(k, c, mu)
    ....:     print("weight {}".format(mu))
    ....:     print(T.list())
    weight [3, 3]
    []
    weight [3, 2, 1]
    [[[1, 1, 1, 2, 2], [2, 2], [3]]]
    weight [3, 1, 1, 1]
    [[[1, 1, 1, 2, 4], [2, 4], [3]], [[1, 1, 1, 2, 3], [2, 3], [4]]]
    weight [2, 2, 2]
    [[[1, 1, 2, 2, 3], [2, 3], [3]]]
    weight [2, 2, 1, 1]
    [[[1, 1, 2, 2, 4], [2, 4], [3]], [[1, 1, 2, 2, 3], [2, 3], [4]]]
    weight [2, 1, 1, 1, 1]
    [[[1, 1, 3, 4, 5], [2, 5], [3]], [[1, 1, 2, 3, 5], [3, 5], [4]],
     [[1, 1, 2, 3, 4], [3, 4], [5]]]
    weight [1, 1, 1, 1, 1, 1]
    [[[1, 3, 4, 5, 6], [2, 6], [4]], [[1, 2, 4, 5, 6], [3, 6], [4]],
     [[1, 2, 3, 4, 6], [4, 6], [5]], [[1, 2, 3, 4, 5], [4, 5], [6]]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2487::

    sage: Sym = SymmetricFunctions(QQ)
    sage: ks = Sym.kschur(3, t=1)
    sage: h = Sym.homogeneous()
    sage: for mu in Partitions(7, max_part=3):
    ....:     print(h(ks(mu)))
    h[3, 3, 1]
    h[3, 2, 2] - h[3, 3, 1]
    h[3, 2, 1, 1] - h[3, 2, 2]
    h[3, 1, 1, 1, 1] - 2*h[3, 2, 1, 1] + h[3, 3, 1]
    h[2, 2, 2, 1] - h[3, 2, 1, 1] - h[3, 2, 2] + h[3, 3, 1]
    h[2, 2, 1, 1, 1] - 2*h[2, 2, 2, 1] - h[3, 1, 1, 1, 1]
        + 2*h[3, 2, 1, 1] + h[3, 2, 2] - h[3, 3, 1]
    h[2, 1, 1, 1, 1, 1] - 3*h[2, 2, 1, 1, 1] + 2*h[2, 2, 2, 1]
        + h[3, 2, 1, 1] - h[3, 2, 2]
    h[1, 1, 1, 1, 1, 1, 1] - 4*h[2, 1, 1, 1, 1, 1] + 4*h[2, 2, 1, 1, 1]
        + 2*h[3, 1, 1, 1, 1] - 4*h[3, 2, 1, 1] + h[3, 3, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2608::

    sage: ks6 = Sym.kschur(6,t=1)
    sage: ks6(h[4,3,1])
    ks6[4, 3, 1] + ks6[4, 4] + ks6[5, 2, 1] + 2*ks6[5, 3]
      + ks6[6, 1, 1] + ks6[6, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2617::

    sage: Sym = SymmetricFunctions(QQ)
    sage: ks = Sym.kschur(3,t=1)
    sage: ks.realization_of()
    3-bounded Symmetric Functions over Rational Field with t=1
    sage: s = Sym.schur()
    sage: s.realization_of()
    Symmetric Functions over Rational Field

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2658::

    sage: k = 6
    sage: weight = Partition([4,3,1])
    sage: for la in Partitions(weight.size(), max_part = k):
    ....:     if la.dominates(weight):
    ....:         print(la)
    ....:         T = WeakTableaux(k, la, weight, representation = 'bounded')
    ....:         print(T.list())
    [6, 2]
    [[[1, 1, 1, 1, 2, 2], [2, 3]]]
    [6, 1, 1]
    [[[1, 1, 1, 1, 2, 2], [2], [3]]]
    [5, 3]
    [[[1, 1, 1, 1, 3], [2, 2, 2]], [[1, 1, 1, 1, 2], [2, 2, 3]]]
    [5, 2, 1]
    [[[1, 1, 1, 1, 2], [2, 2], [3]]]
    [4, 4]
    [[[1, 1, 1, 1], [2, 2, 2, 3]]]
    [4, 3, 1]
    [[[1, 1, 1, 1], [2, 2, 2], [3]]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2740::

    sage: mu = Partition([3,2,1])
    sage: c = mu.to_core(3)
    sage: w = c.to_grassmannian()
    sage: w.stanley_symmetric_function()
    4*m[1, 1, 1, 1, 1, 1] + 3*m[2, 1, 1, 1, 1] + 2*m[2, 2, 1, 1]
        + m[2, 2, 2] + 2*m[3, 1, 1, 1] + m[3, 2, 1]
    sage: sorted(w.reduced_words())
    [[0, 2, 3, 2, 1, 0],
     [0, 3, 2, 1, 3, 0],
     [0, 3, 2, 3, 1, 0],
     [2, 0, 3, 2, 1, 0]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2752::

    sage: Sym = SymmetricFunctions(QQ)
    sage: Q3 = Sym.kBoundedQuotient(3,t=1)
    sage: F3 = Q3.affineSchur()
    sage: m = Q3.kmonomial()
    sage: m(F3([3,2,1]))
    4*m3[1, 1, 1, 1, 1, 1] + 3*m3[2, 1, 1, 1, 1] + 2*m3[2, 2, 1, 1]
        + m3[2, 2, 2] + 2*m3[3, 1, 1, 1] + m3[3, 2, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2799::

    sage: Sym = SymmetricFunctions(QQ)
    sage: Q3 = Sym.kBoundedQuotient(3,t=1)
    sage: F3 = Q3.affineSchur()
    sage: h = Sym.homogeneous()
    sage: f = F3[3,2,1]*h[1]; f
    F3[3, 1, 1, 1, 1] + 3*F3[3, 2, 1, 1] + F3[3, 2, 2] + 2*F3[3, 3, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2810::

    sage: c = Partition([3,2,1]).to_core(3)
    sage: for p in sorted(f.support()):   # Sorted for consistant doctest ordering
    ....:   print("{} {}".format(p, SkewPartition([p.to_core(3).to_partition(),c.to_partition()])))
    [3, 1, 1, 1, 1] [[5, 2, 1, 1, 1], [5, 2, 1]]
    [3, 2, 1, 1] [[6, 3, 1, 1], [5, 2, 1]]
    [3, 2, 2] [[5, 2, 2], [5, 2, 1]]
    [3, 3, 1] [[7, 4, 1], [5, 2, 1]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2976::

sage: T = StrongTableau([[-1,-1,-2,-3],[-2,3,-3,4],[2,3],[-3,-4]], 3)
sage: T.to_transposition_sequence()
[[-2, -1], [3, 4], [0, 2], [-3, -2], [2, 3], [-1, 0], [1, 2], [0, 1]]
sage: T.intermediate_shapes()
[[], [2], [3, 1, 1], [4, 3, 2, 1], [4, 4, 2, 2]]
sage: [T.content_of_marked_head(v+1) for v in range(8)]
[0, 1, -1, 2, -3, 1, 3, -2]
sage: T.left_action([0,1])
[[-1, -1, -2, -3, 5], [-2, 3, -3, 4], [2, 3, -5], [-3, -4], [5]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 2999::

sage: ST = StrongTableaux(3, [6,3,1,1], [4,2,1]); ST
Set of strong 3-tableaux of shape [6, 3, 1, 1] and of weight (4, 2, 1)
sage: ST.list()
[[[-1, -1, -1, -1, 2, 2], [1, -2, -2], [-3], [3]],
 [[-1, -1, -1, -1, 2, -2], [1, -2, 2], [-3], [3]],
 [[-1, -1, -1, -1, -2, -2], [1, 2, 2], [-3], [3]],
 [[-1, -1, -1, -1, 2, 3], [1, -2, 3], [-2], [-3]],
 [[-1, -1, -1, -1, 2, 3], [1, -2, -3], [-2], [3]],
 [[-1, -1, -1, -1, 2, -3], [1, -2, 3], [-2], [3]],
 [[-1, -1, -1, -1, -2, 3], [1, 2, 3], [-2], [-3]],
 [[-1, -1, -1, -1, -2, 3], [1, 2, -3], [-2], [3]],
 [[-1, -1, -1, -1, -2, -3], [1, 2, 3], [-2], [3]]]
sage: ks = SymmetricFunctions(QQ).kschur(3,1)
sage: m = SymmetricFunctions(QQ).m()
sage: m(ks[3,2,1,1]).coefficient([4,2,1])
9

Sage example in ./kschurnotes/notes-mike-anne.tex, line 3243::

  sage: W = WeylGroup(['A',3,1])
  sage: [w.reduced_word() for w in W.pieri_factors()]
  [[], [0], [1], [2], [3], [1, 0], [2, 0], [0, 3], [2, 1], [3, 1], [3, 2],
   [2, 1, 0], [1, 0, 3], [0, 3, 2], [3, 2, 1]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 3251::

  sage: A = NilCoxeterAlgebra(WeylGroup(['A',3,1]), prefix = 'A')
  sage: A.homogeneous_noncommutative_variables([2])
  A[1,0] + A[2,0] + A[0,3] + A[3,2] + A[3,1] + A[2,1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 3258::

  sage: A.k_schur_noncommutative_variables([2,2])
  A[0,3,1,0] + A[3,1,2,0] + A[1,2,0,1] + A[3,2,0,3] + A[2,0,3,1] + A[2,3,1,2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 3265::

  sage: Sym = SymmetricFunctions(ZZ)
  sage: ks = Sym.kschur(5,t=1)
  sage: ks[2,1]*ks[2,1]
  ks5[2, 2, 1, 1] + ks5[2, 2, 2] + ks5[3, 1, 1, 1] + 2*ks5[3, 2, 1]
  + ks5[3, 3] + ks5[4, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 3536::

    sage: la = Partition([3,2,1,1])
    sage: la.k_atom(4)
    [[[1, 1, 1, 4], [2, 2], [3]], [[1, 1, 1], [2, 2], [3], [4]]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 3639::

    sage: s = SymmetricFunctions(QQ['t']).schur()
    sage: G1 = s[1]
    sage: G211 = G1.hl_creation_operator([2,1]); G211
    s[2, 1, 1] + t*s[2, 2] + t*s[3, 1]
    sage: G3211 = G211.hl_creation_operator([3]); G3211
    s[3, 2, 1, 1] + t*s[3, 2, 2] + t*s[3, 3, 1] + t*s[4, 1, 1, 1]
     + (2*t^2+t)*s[4, 2, 1] + t^2*s[4, 3] + (t^3+t^2)*s[5, 1, 1]
     + 2*t^3*s[5, 2] + t^4*s[6, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 3954::

  sage: T = WeakTableau([[1,1,2,3,4,4,5,5,6],[2,3,5,5,6],[3,4,7],
  ....:  [5,6],[6],[7]],4)
  sage: T.k_charge()
  12

Sage example in ./kschurnotes/notes-mike-anne.tex, line 3962::

  sage: Sym = SymmetricFunctions(QQ['t'].fraction_field())
  sage: Qp = Sym.hall_littlewood().Qp()
  sage: ks = Sym.kBoundedSubspace(3).kschur()
  sage: t = ks.base_ring().gen()
  sage: ks(Qp[3,2,2,1])
  ks3[3, 2, 2, 1] + t*ks3[3, 3, 1, 1] + t^2*ks3[3, 3, 2]
  sage: sum(t^T.k_charge()*ks(la) for la in Partitions(8, max_part=3)
  ....:  for T in WeakTableaux(3,la,[3,2,2,1],representation = 'bounded'))
  ks3[3, 2, 2, 1] + t*ks3[3, 3, 1, 1] + t^2*ks3[3, 3, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4055::

    sage: t = var('t')                                                                  # needs sage.symbolic
    sage: for mu in Partitions(5):                                                      # needs sage.symbolic
    ....:     print("{} {}".format(mu, sum(t^T.spin() for T in StrongTableaux(3,[4,1,1],mu))))
    [5] 0
    [4, 1] t
    [3, 2] t
    [3, 1, 1] 2*t + 1
    [2, 2, 1] 2*t + 1
    [2, 1, 1, 1] 3*t + 3
    [1, 1, 1, 1, 1] 4*t + 6
    sage: StrongTableaux( 3, [4,1,1], (1,)*5 ).cardinality()
    10
    sage: StrongTableaux( 3, [4,1,1], (1,)*5 ).list()
    [[[-1, -2, -3, 4], [-4], [-5]],
     [[-1, -2, -3, -4], [4], [-5]],
     [[-1, -2, -3, -5], [-4], [4]],
     [[-1, -2, 4, -4], [-3], [-5]],
     [[-1, -2, 4, -5], [-3], [-4]],
     [[-1, -2, -4, -5], [-3], [4]],
     [[-1, -3, 4, -4], [-2], [-5]],
     [[-1, -3, 4, -5], [-2], [-4]],
     [[-1, -3, -4, -5], [-2], [4]],
     [[-1, 4, -4, -5], [-2], [-3]]]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4385::

    sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
    sage: ks4 = Sym.kschur(4)
    sage: ks4([3, 1, 1]).hl_creation_operator([1])
    (t-1)*ks4[2, 2, 1, 1] + t^2*ks4[3, 1, 1, 1] + t^3*ks4[3, 2, 1]
     + (t^3-t^2)*ks4[3, 3] + t^4*ks4[4, 1, 1]
    sage: ks4([3, 1, 1]).hl_creation_operator([2])
    t*ks4[3, 2, 1, 1] + t^2*ks4[3, 3, 1] + t^2*ks4[4, 1, 1, 1]
     + t^3*ks4[4, 2, 1]
    sage: ks4([3, 1, 1]).hl_creation_operator([3])
    ks4[3, 3, 1, 1] + t*ks4[4, 2, 1, 1] + t^2*ks4[4, 3, 1]
    sage: ks4([3, 1, 1]).hl_creation_operator([4])
    ks4[4, 3, 1, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4456::

    sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
    sage: ks3 = Sym.kschur(3)
    sage: ks3([3,2]).omega()
    Traceback (most recent call last):
    ...
    ValueError: t^2*s[1, 1, 1, 1, 1] + t*s[2, 1, 1, 1] + s[2, 2, 1] is not
    in the image

    sage: s = Sym.schur()
    sage: s(ks3[3,2])
    s[3, 2] + t*s[4, 1] + t^2*s[5]
    sage: t = s.base_ring().gen()
    sage: invert = lambda x: s.base_ring()(x.subs(t=1/t))
    sage: ks3(s(ks3([3,2])).omega().map_coefficients(invert))
    1/t^2*ks3[1, 1, 1, 1, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4478::

    sage: ks3[3,2].omega_t_inverse()
    1/t^2*ks3[1, 1, 1, 1, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4686::

    sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
    sage: ks3 = Sym.kschur(3)
    sage: ks3[3,1].coproduct()
    ks3[] # ks3[3, 1] + ks3[1] # ks3[2, 1] + (t+1)*ks3[1] # ks3[3]
    + ks3[1, 1] # ks3[2] + ks3[2] # ks3[1, 1] + (t+1)*ks3[2] # ks3[2]
    + ks3[2, 1] # ks3[1] + (t+1)*ks3[3] # ks3[1]  + ks3[3, 1] # ks3[]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4720::

    sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
    sage: ks2 = Sym.kschur(2)
    sage: ks3 = Sym.kschur(3)
    sage: ks5 = Sym.kschur(5)
    sage: ks5(ks3[2])*ks5(ks2[1])
    ks5[2, 1] + ks5[3]
    sage: ks5(ks3[2])*ks5(ks2[2,1])
    ks5[2, 2, 1] + ks5[3, 1, 1] + (t+1)*ks5[3, 2] + (t+1)*ks5[4, 1]
      + t*ks5[5]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4779::

    sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
    sage: ks3 = Sym.kschur(3)
    sage: ks4 = Sym.kschur(4)
    sage: ks5 = Sym.kschur(5)
    sage: ks4(ks3[3,2,1,1])
    ks4[3, 2, 1, 1] + t*ks4[3, 3, 1] + t*ks4[4, 1, 1, 1] + t^2*ks4[4, 2, 1]
    sage: ks5(ks3[3,2,1,1])
    ks5[3, 2, 1, 1] + t*ks5[3, 3, 1] + t*ks5[4, 1, 1, 1] + t^2*ks5[4, 2, 1]
     + t^2*ks5[4, 3] + t^3*ks5[5, 1, 1]

    sage: ks5(ks4[3,2,1,1])
    ks5[3, 2, 1, 1]
    sage: ks5(ks4[4,3,3,2,1,1])
    ks5[4, 3, 3, 2, 1, 1] + t*ks5[4, 4, 3, 1, 1, 1]
     + t^2*ks5[5, 3, 3, 1, 1, 1]
    sage: ks5(ks4[4,3,3,2,1,1,1])
    ks5[4, 3, 3, 2, 1, 1, 1] + t*ks5[4, 3, 3, 3, 1, 1]
     + t*ks5[4, 4, 3, 1, 1, 1, 1] + t^2*ks5[4, 4, 3, 2, 1, 1]
     + t^2*ks5[5, 3, 3, 1, 1, 1, 1] + t^3*ks5[5, 3, 3, 2, 1, 1]
     + t^4*ks5[5, 4, 3, 1, 1, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4858::

    sage: Sym = SymmetricFunctions(FractionField(QQ['q,t']))
    sage: H = Sym.macdonald().H()
    sage: ks = Sym.kschur(3)
    sage: ks(H[3])
    q^3*ks3[1, 1, 1] + (q^2+q)*ks3[2, 1] + ks3[3]
    sage: ks(H[3,2])                          # long time (2.11 s, 2013)
    q^4*ks3[1, 1, 1, 1, 1] + (q^3*t+q^3+q^2)*ks3[2, 1, 1, 1]
     + (q^3*t+q^2*t+q^2+q)*ks3[2, 2, 1]
     + (q^2*t+q*t+q)*ks3[3, 1, 1] + ks3[3, 2]
    sage: ks(H[3,1,1])
    q^3*ks3[1, 1, 1, 1, 1] + (q^3*t^2+q^2+q)*ks3[2, 1, 1, 1]
     + (q^2*t^2+q^2*t+q*t+q)*ks3[2, 2, 1]
     + (q^2*t^2+q*t^2+1)*ks3[3, 1, 1] + t*ks3[3, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 4965::

    sage: Sym = SymmetricFunctions(QQ)
    sage: Q3 = Sym.kBoundedQuotient(3,t=1)
    sage: F = Q3.affineSchur()
    sage: p = Sym.power()
    sage: F[2,1]*p[2]
    -F3[1, 1, 1, 1, 1] - F3[2, 1, 1, 1] + F3[3, 1, 1] + F3[3, 2]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 5028::

    sage: R = QQ[I]; z4 = R.zeta(4)
    sage: Sym = SymmetricFunctions(R)
    sage: ks3z = Sym.kschur(3,t=z4)
    sage: ks3 = Sym.kschur(3,t=1)
    sage: p = Sym.p()
    sage: p(ks3z[2, 2, 2, 2, 2, 2, 2, 2])     # long time (17s on sage.math, 2013)
    1/12*p[4, 4, 4, 4] + 1/4*p[8, 8] - 1/3*p[12, 4]
    sage: p(ks3[2,2])
    1/12*p[1, 1, 1, 1] + 1/4*p[2, 2] - 1/3*p[3, 1]
    sage: p(ks3[2,2]).plethysm(p[4])
    1/12*p[4, 4, 4, 4] + 1/4*p[8, 8] - 1/3*p[12, 4]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 5042::

    sage: ks3z[3, 3, 3, 3]*ks3z[2, 1]   # long time (10s on sage.math, 2013)
    ks3[3, 3, 3, 3, 2, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 5197::

    sage: Sym = SymmetricFunctions(QQ['t'].fraction_field())
    sage: Q3 = Sym.kBoundedQuotient(3)
    sage: dks = Q3.dual_k_Schur()
    sage: dks[2, 1, 1]*dks[3, 2, 1]          # long time (25.7s, 2013)
    (t^7+t^6)*dks3[2, 1, 1, 1, 1, 1, 1, 1, 1]
    + (t^4+t^3+t^2)*dks3[2, 2, 2, 1, 1, 1, 1]
    + (t^3+t^2)*dks3[2, 2, 2, 2, 1, 1]
    + (t^5+2*t^4+2*t^3+t^2)*dks3[2, 2, 2, 2, 2]
    + (t^5+2*t^4+t^3)*dks3[3, 1, 1, 1, 1, 1, 1, 1]
    + (2*t^5+3*t^4+4*t^3+3*t^2+t)*dks3[3, 2, 1, 1, 1, 1, 1]
    + (2*t^2+t+1)*dks3[3, 2, 2, 1, 1, 1]
    + (t^4+3*t^3+4*t^2+3*t+1)*dks3[3, 2, 2, 2, 1]
    + (t^5+t^4+4*t^3+4*t^2+3*t+1)*dks3[3, 3, 1, 1, 1, 1]
    + (2*t^5+3*t^4+5*t^3+6*t^2+4*t+2)*dks3[3, 3, 2, 1, 1]
    + (t^4+t^3+3*t^2+2*t+1)*dks3[3, 3, 2, 2]
    + (t^5+3*t^4+3*t^3+4*t^2+2*t+1)*dks3[3, 3, 3, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 5294::

    sage: ks2 = SymmetricFunctions(QQ['t']).kschur(2)
    sage: HLQp = SymmetricFunctions(QQ['t']).hall_littlewood().Qp()
    sage: ks2( (HLQp(ks2[1,1])*HLQp(ks2[1])).restrict_parts(2) )
    ks2[1, 1, 1] + (-t+1)*ks2[2, 1]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 5303::

    sage: dks = SymmetricFunctions(QQ['t']).kBoundedQuotient(2).dks()
    sage: dks[2,1].coproduct()
    dks2[] # dks2[2, 1] + (-t+1)*dks2[1] # dks2[1, 1] +
     dks2[1] # dks2[2] + (-t+1)*dks2[1, 1] # dks2[1] +
     dks2[2] # dks2[1] + dks2[2, 1] # dks2[]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 5560::

sage: Sym = SymmetricFunctions(QQ)
sage: Sym3 = Sym.kBoundedSubspace(3,t=1)
sage: Kks3 = Sym3.K_kschur()
sage: s = Sym.s()
sage: m = Sym.m()
sage: s(Kks3[3,1])
s[3] + s[3, 1] + s[4]
sage: m(Kks3[3,1])
m[1, 1, 1] + 4*m[1, 1, 1, 1] + m[2, 1] + 3*m[2, 1, 1] + 2*m[2, 2]
+ m[3] + 2*m[3, 1] + m[4]
sage: ks3 = Sym3.kschur()
sage: ks3(Kks3[3,1])
ks3[3] + ks3[3, 1]
sage: Kks3[3,1]*Kks3[2]            # long time (11.85 s, 2013)
-Kks3[3, 1, 1] - Kks3[3, 2] + Kks3[3, 2, 1] + Kks3[3, 3]
sage: Kks3[3,1].coproduct()
Kks3[] # Kks3[3, 1] - Kks3[1] # Kks3[2] + Kks3[1] # Kks3[2, 1]
+ 2*Kks3[1] # Kks3[3] + Kks3[1, 1] # Kks3[2] - Kks3[2] # Kks3[1]
+ Kks3[2] # Kks3[1, 1] + 2*Kks3[2] # Kks3[2] + Kks3[2, 1] # Kks3[1]
+ 2*Kks3[3] # Kks3[1] + Kks3[3, 1] # Kks3[]

Sage example in ./kschurnotes/notes-mike-anne.tex, line 5588::

sage: # long time
sage: SymQ3 = Sym.kBoundedQuotient(3,t=1)
sage: G1 = SymQ3.AffineGrothendieckPolynomial([1],6)
sage: G2 = SymQ3.AffineGrothendieckPolynomial([2],6)
sage: (G1*G2).lift().scalar(Kks3[3,1])
-1
"""



# sage_setup: distribution = sagemath-repl
r"""
This file contains some tests that Sage command line options actually
do something.

We test the following command line options:

test.py
/path/to/test.py
test.sage
/path/to/test.sage
test.spyx
/path/to/test.spyx
--advanced
-c
--cython
--dev
--ecl
--experimental
--fixdoctests
--gap
--gdb
--gp
-h
--help
--info
--ipython
--lisp
--maxima
--min
--mwrank
--optional
--preparse
--python
--python3
-q
--R
--root
--rst2ipynb
--ipynb2rst
--sh
--singular
--sqlite3
--standard
--startuptime
-t
-v
--zzfoobar (illegal option)


AUTHORS:

- Jeroen Demeyer (2010-11-20): initial version (:issue:`10300`)


TESTS:

Run Sage itself with various options::

    sage: from sage.tests.fixtures import check_executable

    sage: # long time
    sage: from sage.version import banner
    sage: (out, err, ret) = check_executable([
    ....:     "sage"], pydebug_ignore_warnings=True)
    sage: out.find(banner) >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # long time
    sage: (out, err, ret) = check_executable([
    ....:     "sage"], "3^33\n", pydebug_ignore_warnings=True)
    sage: out.find(banner) >= 0
    True
    sage: out.find("5559060566555523") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # long time
    sage: (out, err, ret) = check_executable([
    ....:     "sage", "-q"], "3^33\n", pydebug_ignore_warnings=True)
    sage: out.find(banner) >= 0
    False
    sage: out.find("5559060566555523") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # long time
    sage: (out, err, ret) = check_executable([
    ....:     "sage", "-c", "print(3^33)"])
    sage: print(out)
    5559060566555523
    sage: err
    ''
    sage: ret
    0

    sage: # long time
    sage: (out, err, ret) = check_executable([
    ....:     "sage", "--min", "-c", "print(3^33)"])
    sage: print(out)
    5559060566555523
    sage: err
    ''
    sage: ret
    0

    sage: # long time
    sage: (out, err, ret) = check_executable([
    ....:     "sage", "--startuptime"])
    sage: out.find("Slowest module import") >= 0
    True
    sage: err
    ''
    sage: ret
    0

Test help::

    sage: (out, err, ret) = check_executable(["sage", "-h"])
    sage: out.find("evaluate cmd as sage") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: (out, err, ret) = check_executable(["sage", "--help"])
    sage: out.find("evaluate cmd as sage") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: (out, err, ret) = check_executable(["sage", "--advanced"])
    sage: out.find("run the Sage cleaner.") >= 0
    True
    sage: err
    ''
    sage: ret
    0
    sage: out.find("print the Sage root directory") >= 0 # optional - sage_spkg
    True
    sage: out.find("regular expression search through the Sage") >= 0 # optional - sage_spkg
    True

Basic information about the Sage installation::

    sage: (out, err, ret) = check_executable(["sage", "-v"])
    sage: from sage.version import banner
    sage: out.find(banner) >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # optional - sage_spkg
    sage: (out, err, ret) = check_executable(["sage", "--root"])
    sage: len(out) >= 2
    True
    sage: err
    ''
    sage: ret
    0

Test ``sage --info [packages]``::

    sage: # optional - sage_spkg
    sage: out, err, ret = check_executable(["sage", "--info", "sqlite"])
    sage: print(out)
    sqlite...
    SQLite is a software library that implements a self-contained,
    serverless, zero-configuration, transactional SQL database engine.
    ...
    sage: err
    ''
    sage: ret
    0

Test ``sage-run`` on a Python file, both with an absolute and with a relative path::

    sage: import tempfile
    sage: with tempfile.TemporaryDirectory() as dir:
    ....:     name = 'python_test_file.py'
    ....:     fullname = os.path.join(dir, name)
    ....:     with open(fullname, 'w') as F:
    ....:         _ = F.write("print(3^33)\n")
    ....:     check_executable(["sage", fullname])
    ....:     check_executable(["sage", name], cwd=dir)
    ('34\n', '', 0)
    ('34\n', '', 0)

The same as above, but now with a ``.sage`` file.  This indirectly
also tests the preparser::

    sage: import tempfile
    sage: with tempfile.TemporaryDirectory() as dir:  # long time
    ....:     name = 'sage_test_file.sage'
    ....:     fullname = os.path.join(dir, name)
    ....:     with open(fullname, 'w') as F:
    ....:         _ = F.write("k.<a> = GF(5^3); print(a^124)\n")
    ....:     check_executable(["sage", fullname])
    ....:     check_executable(["sage", name], cwd=dir)
    ('1\n', '', 0)
    ('1\n', '', 0)

Test running a ``.spyx`` file::

    sage: import tempfile
    sage: with tempfile.TemporaryDirectory() as dir:  # long time
    ....:     name = 'sage_test_file.spyx'
    ....:     fullname = os.path.join(dir, name)
    ....:     with open(fullname, 'w') as F:
    ....:         _ = F.write("from cysignals.signals cimport *\n")
    ....:         _ = F.write("from sage.rings.integer cimport Integer\n")
    ....:         _ = F.write("cdef long i, s = 0\n")
    ....:         _ = F.write("sig_on()\n")
    ....:         _ = F.write("for i in range(5): s += i\n")
    ....:         _ = F.write("sig_off()\n")
    ....:         _ = F.write("print(Integer(s))")
    ....:     check_executable(["sage", fullname], pydebug_ignore_warnings=True)
    ....:     check_executable(["sage", name], cwd=dir, pydebug_ignore_warnings=True)
    ('10\n', 'Compiling .../sage_test_file.spyx...\n', 0)
    ('10\n', 'Compiling sage_test_file.spyx...\n', 0)

Testing ``sage --preparse FILE`` and ``sage -t FILE``.  First create
a file and preparse it::

    sage: s = "# -*- coding: utf-8 -*-\n'''This is a test file.\nAnd I am its doctest'''\ndef my_add(a):\n    '''\n    Add 2 to a.\n\n        EXAMPLES::\n\n            sage: my_add(2)\n            4\n        '''\n    return a + 2\n"
    sage: script = os.path.join(tmp_dir(), 'my_script.sage')
    sage: script_py = script + '.py'
    sage: F = open(script, 'w')
    sage: _ = F.write(s)
    sage: F.close()
    sage: (out, err, ret) = check_executable(["sage", "--preparse", script])
    sage: ret
    0
    sage: os.path.isfile(script_py)
    True

Now test my_script.sage and the preparsed version my_script.sage.py::

    sage: # long time
    sage: (out, err, ret) = check_executable([
    ....:     "sage", "-t", "--optional=sage", script])
    sage: ret
    0
    sage: out.find("All tests passed!") >= 0
    True
    sage: (out, err, ret) = check_executable([
    ....:     "sage", "-t", "--optional=sage", script_py])
    sage: ret
    0
    sage: out.find("All tests passed!") >= 0
    True

Test that the coding line and doctest are preserved::

    sage: Fpy = open(script_py, "r")
    sage: Fpy.readline()
    '# -*- coding: utf-8 -*-\n'
    sage: Fpy.readline()
    "'''This is a test file.\n"
    sage: Fpy.readline()
    "And I am its doctest'''\n"

Now for a file which should fail tests::

    sage: s = s.replace('4', '5') # (2+2 != 5)
    sage: F = open(script, 'w')
    sage: _ = F.write(s)
    sage: F.close()
    sage: (out, err, ret) = check_executable([           # long time
    ....:     "sage", "-t", "--optional=sage", script])
    sage: ret                                           # long time
    1
    sage: out.find("1 item had failures:") >= 0         # long time
    True

Test ``sage -t --debug -p 2`` on a ReST file, the ``-p 2`` should
be ignored. In Pdb, we run the ``help`` command::

    sage: s = "::\n\n    sage: assert True is False\n    sage: 2 + 2\n    5"
    sage: script = tmp_filename(ext='.rst')
    sage: F = open(script, 'w')
    sage: _ = F.write(s)
    sage: F.close()
    sage: (out, err, ret) = check_executable([           # long time
    ....:     "sage", "-t", "--optional=sage", "--debug",
    ....:     "-p", "2", "--warn-long", "0", script], "help")
    sage: print(out)                                    # long time
    Debugging requires single-threaded operation, setting number of threads to 1.
    Running doctests with ID...
    Doctesting 1 file.
    ...
    **********************************************************************
    File "...", line 3, in ...
    Failed example:
        assert True is False
    Exception raised:
        Traceback (most recent call last):
        ...
        AssertionError
    > <doctest ...>(1)<module>()
    -> assert True is False
    (Pdb)
    Documented commands (type help <topic>):
    ========================================
    ...
    **********************************************************************
    File "...", line 4, in ...
    Failed example:
        2 + 2
    Expected:
        5
    Got:
        4
    **********************************************************************
    Previously executed commands:
        s...: assert True is False
    sage:
    <BLANKLINE>
    Returning to doctests...
    **********************************************************************
    1 item had failures:
       2 of   3 in ...
        [2 tests, 2 failures, ...]
    ...
    sage: ret                                           # long time
    1

Now run a test for the fixdoctests script and, in particular, check that the
issues raised in :issue:`10589` are fixed. We have to go to slightly silly
lengths to doctest the output.::

    sage: test='r\"\"\"Add a doc-test for the fixdoctest command line option and, in particular, check that\n:issue:`10589` is fixed.\n\nEXAMPLES::\n\n    sage: 1+1              # incorrect output\n    3\n    sage: m=matrix(ZZ,3)   # output when none is expected\n    [0 0 0]\n    [0 0 0]\n    [1 0 0]\n    sage: (2/3)*m          # no output when it is expected\n    sage: mu=PartitionTuple([[4,4],[3,3,2,1],[1,1]])   # output when none is expected\n    [4, 4, 3, 3, 2, 1, 1]\n    sage: mu.pp()          # uneven indentation\n    ****\n    ****\n    sage: PartitionTuples.options(convention="French")\n    sage: mu.pp()         # fix doctest with uneven indentation\n    sage: PartitionTuples.options._reset()\n\"\"\"\n'
    sage: test_file = os.path.join(tmp_dir(), 'test_file.py')
    sage: F = open(test_file, 'w')
    sage: _ = F.write(test)
    sage: F.close()
    sage: (out, err, ret) = check_executable(["sage", "--fixdoctests", test_file])       # long time
    sage: with open(test_file, 'r') as f:                                               # long time
    ....:     fixed_test = f.read()
    sage: import difflib                                                                # long time
    sage: list(difflib.unified_diff(test.splitlines(), fixed_test.splitlines()))[2:-1]  # long time
    ['@@ -4,18 +4,23 @@\n',
     ' EXAMPLES::',
     ' ',
     '     sage: 1+1              # incorrect output',
     '-    3',
     '+    2',
     '     sage: m=matrix(ZZ,3)   # output when none is expected',
     '+    sage: (2/3)*m          # no output when it is expected',
     '     [0 0 0]',
     '     [0 0 0]',
     '-    [1 0 0]',
     '-    sage: (2/3)*m          # no output when it is expected',
     '+    [0 0 0]',
     '     sage: mu=PartitionTuple([[4,4],[3,3,2,1],[1,1]])   # output when none is expected',
     '-    [4, 4, 3, 3, 2, 1, 1]',
     '     sage: mu.pp()          # uneven indentation',
     '-    ****',
     '-    ****',
     '+       ****   ***   *',
     '+       ****   ***   *',
     '+              **',
     '+              *',
     '     sage: PartitionTuples.options(convention="French")',
     '     sage: mu.pp()         # fix doctest with uneven indentation',
     '+    *',
     '+    **',
     '+    ****   ***   *',
     '+    ****   ***   *',
     '     sage: PartitionTuples.options._reset()']

Test external programs being called by Sage::

    sage: (out, err, ret) = check_executable(["sage", "--sh"], "echo Hello World\nexit 42\n")
    sage: out.find("Hello World\n") >= 0
    True
    sage: ret
    42

    sage: (out, err, ret) = check_executable(["sage", "--sh", "-c", "echo Hello World; exit 42"])
    sage: out.find("Hello World\n") >= 0
    True
    sage: ret
    42

    sage: # long time
    sage: (out, err, ret) = check_executable([
    ....:     "sage", "--ipython"], "\n3**33\n", pydebug_ignore_warnings=True)
    sage: out.find("5559060566555523") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: (out, err, ret) = check_executable(["sage", "--python"], "print(3^33)\n")
    sage: out
    '34\n'
    sage: err
    ''
    sage: ret
    0

    sage: (out, err, ret) = check_executable(["sage", "--python3"], "print(3^33)\n")
    sage: out
    '34\n'
    sage: err
    ''
    sage: ret
    0

    sage: (out, err, ret) = check_executable(["sage", "--cython"])
    sage: print(err)
    ...
    cython: error: cython: Need at least one source file

    sage: # needs sage.symbolic
    sage: def has_tty():
    ....:     try:
    ....:         os.open(os.ctermid(), os.O_RDONLY)
    ....:         return True
    ....:     except OSError:
    ....:         return False
    sage: (out, err, ret) = check_executable(["sage", "--ecl"], "(* 12345 54321)\n")
    sage: out.find("Embeddable Common-Lisp") >= 0
    True
    sage: out.find("670592745") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # needs sage.symbolic
    sage: (out, err, ret) = check_executable(["sage", "--lisp"], "(* 12345 54321)\n")
    sage: out.find("Embeddable Common-Lisp") >= 0
    True
    sage: out.find("670592745") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # needs sage.libs.gap
    sage: # long time
    sage: (out, err, ret) = check_executable([
    ....:     "sage", "--gap", "-q"], "Size(SymmetricGroup(5));\n")
    sage: out
    '120\n'
    sage: err.replace('gap: halving pool size.', '').strip()
    ''
    sage: ret
    0

    sage: (out, err, ret) = check_executable([                   # long time  # optional - gdb
    ....:     "sage", "--gdb"], 'quit\n')
    sage: out.find('(gdb) ') >= 0                               # long time  # optional - gdb
    True
    sage: ret                                                   # long time  # optional - gdb
    0

    sage: # needs sage.libs.eclib
    sage: (out, err, ret) = check_executable(["sage", "--mwrank", "-v0", "-q"], "0 0 0 0 1\n")
    sage: out
    'Curve [0,0,0,0,1] :\tRank = 0\n\n'
    sage: err
    ''
    sage: ret
    0

    sage: # needs sage.libs.singular
    sage: (out, err, ret) = check_executable(["sage", "--singular"], "12345*54321;\n")
    sage: out.find("A Computer Algebra System for Polynomial Computations") >= 0
    True
    sage: out.find("670592745") >= 0
    True
    sage: err
    ''
    sage: ret
    0

Test GP using the ``-f`` option which prevents the reading of a ``.gprc``
configuration file::

    sage: # needs sage.libs.pari
    sage: (out, err, ret) = check_executable(["sage", "--gp", "-f"], "3^33\nquit(42)\n")
    sage: out.find("PARI/GP") >= 0
    True
    sage: out.find("5559060566555523") >= 0
    True
    sage: err
    ''
    sage: ret
    42

Some programs of which we check functionality using only ``--version``::

    sage: # needs sage.symbolic
    sage: (out, err, ret) = check_executable(["sage", "--maxima", "--version"])
    sage: out.find("Maxima ") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # optional - r
    sage: (out, err, ret) = check_executable(["sage", "--R", "--version"])
    sage: out.find("R version ") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: (out, err, ret) = check_executable(["sage", "--sqlite3", "--version"])
    sage: out.startswith("3.")
    True
    sage: err
    ''
    sage: ret
    0

Check some things requiring an internet connection::

    sage: # optional - internet
    sage: (out, err, ret) = check_executable(["sage", "--standard"])
    sage: out.find("cython") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # optional - internet
    sage: (out, err, ret) = check_executable(["sage", "--optional"])
    sage: out.find("database_cremona_ellcurve") >= 0
    True
    sage: err
    ''
    sage: ret
    0

    sage: # optional - internet
    sage: (out, err, ret) = check_executable(["sage", "--experimental"])
    sage: out.find("valgrind") >= 0
    True
    sage: err
    ''
    sage: ret
    0

Check an illegal command line option.  This outputs an error to stdout,
but we allow stderr in case this changes in the future::

    sage: (out, err, ret) = check_executable(["sage", "--zzfoobar"])
    sage: (out+err).find("unknown option: --zzfoobar") >= 0
    True
    sage: ret > 0
    True

Test ``sage --rst2ipynb file.rst`` on a ReST file::

    sage: s = "::\n\n    sage: 2^10\n    1024\n    sage: 2 + 2\n    4"
    sage: input = tmp_filename(ext='.rst')
    sage: with open(input, 'w') as F:
    ....:     _ = F.write(s)
    sage: L = ["sage", "--rst2ipynb", input]
    sage: (out, err, ret) = check_executable(L)           # optional - rst2ipynb
    sage: err                                            # optional - rst2ipynb
    ''
    sage: ret                                            # optional - rst2ipynb
    0
    sage: from json import loads                         # optional - rst2ipynb
    sage: d = loads(out)                                 # optional - rst2ipynb
    sage: sorted(d.keys())                               # optional - rst2ipynb
    ['cells', 'metadata', 'nbformat', 'nbformat_minor']
    sage: d['cells'][1]['source']                        # optional - rst2ipynb
    ['2^10']
    sage: d['cells'][2]['source']                        # optional - rst2ipynb
    ['2 + 2']

Test ``sage --rst2ipynb file.rst file.ipynb`` on a ReST file::

    sage: s = "::\n\n    sage: 2^10\n    1024\n    sage: 2 + 2\n    4"
    sage: input = tmp_filename(ext='.rst')
    sage: output = tmp_filename(ext='.ipynb')
    sage: with open(input, 'w') as F:
    ....:     _ = F.write(s)
    sage: L = ["sage", "--rst2ipynb", input, output]
    sage: check_executable(L)                              # optional - rst2ipynb
    ('', '', 0)
    sage: import json                                     # optional - rst2ipynb
    sage: d = json.load(open(output,'r'))                 # optional - rst2ipynb
    sage: type(d)                                         # optional - rst2ipynb
    <class 'dict'>
    sage: sorted(d.keys())                                # optional - rst2ipynb
    ['cells', 'metadata', 'nbformat', 'nbformat_minor']
    sage: d['metadata']                                   # optional - rst2ipynb
    {'kernelspec': {'display_name': 'sagemath', 'name': 'sagemath'}}
    sage: d['cells'][1]['cell_type']                      # optional - rst2ipynb
    'code'

Test ``sage --ipynb2rst file.ipynb file.rst`` on a ipynb file::

    sage: s = r'''{
    ....:  "cells": [
    ....:   {
    ....:    "cell_type": "code",
    ....:    "execution_count": 1,
    ....:    "metadata": {},
    ....:    "outputs": [
    ....:     {
    ....:      "data": {
    ....:       "text/plain": [
    ....:        "2"
    ....:       ]
    ....:      },
    ....:      "execution_count": 1,
    ....:      "metadata": {},
    ....:      "output_type": "execute_result"
    ....:     }
    ....:    ],
    ....:    "source": [
    ....:     "1+1"
    ....:    ]
    ....:   },
    ....:   {
    ....:    "cell_type": "code",
    ....:    "execution_count": null,
    ....:    "metadata": {},
    ....:    "outputs": [],
    ....:    "source": []
    ....:   }
    ....:  ],
    ....:  "metadata": {
    ....:   "kernelspec": {
    ....:    "display_name": "SageMath 8.3.beta4",
    ....:    "language": "",
    ....:    "name": "sagemath"
    ....:   },
    ....:   "language_info": {
    ....:    "codemirror_mode": {
    ....:     "name": "ipython",
    ....:     "version": 2
    ....:    },
    ....:    "file_extension": ".py",
    ....:    "mimetype": "text/x-python",
    ....:    "name": "python",
    ....:    "nbconvert_exporter": "python",
    ....:    "pygments_lexer": "ipython2",
    ....:    "version": "2.7.15"
    ....:   }
    ....:  },
    ....:  "nbformat": 4,
    ....:  "nbformat_minor": 2
    ....: }
    ....: '''
    sage: t = '.. escape-backslashes\n.. default-role:: math\n\n\n::\n\n    sage: 1+1\n    2\n\n\n\n\n'
    sage: input = tmp_filename(ext='.ipynb')
    sage: output = tmp_filename(ext='.rst')
    sage: with open(input, 'w') as F:
    ....:     _ = F.write(s)
    sage: L = ["sage", "--ipynb2rst", input, output]
    sage: _ = check_executable(L)                # long time  # optional - pandoc
    sage: print(open(output, 'r').read() == t)  # long time  # optional - pandoc  # known bug #32697
    True
"""



# sage_setup: distribution = sagemath-repl
r"""
Ensure that certain modules are not loaded on startup.

Check that IPython is not imported at startup (:issue:`18726`). It is
imported by the doctest framework, so the simple test like above would
not work. Instead, we test this by starting a new Python process::

    sage: from sage.tests.fixtures import check_executable
    sage: environment = "sage.all"
    sage: cmd = f"from {environment} import *\nprint('IPython' in sys.modules)\n"
    sage: print(check_executable(["python3"], cmd)[0])  # long time
    False

Check that numpy (:issue:`11714`) and pyparsing are not imported on startup
as they increase the startup time. Since :issue:`23696` those are imported
by the doctest framework via a matplotlib import. Again the simple test
would not work (but we don't have to avoid loading IPython)::

    sage: from sage.tests.fixtures import check_executable
    sage: cmd = "print('numpy' in sys.modules)\n"
    sage: print(check_executable(["sage", "-c", cmd])[0])  # long time
    False
    sage: cmd = "print('pyparsing' in sys.modules)\n"
    sage: print(check_executable(["sage", "-c", cmd])[0])  # long time
    False
"""



# sage_setup: distribution = sagemath-repl
"""
IPython Backend for the Sage Rich Output System

This module defines the IPython backends for
:mod:`sage.repl.rich_output`.
"""

# ****************************************************************************
#       Copyright (C) 2015 Volker Braun <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import os
import html
from IPython.display import publish_display_data
from sage.repl.rich_output.backend_base import BackendBase
from sage.repl.rich_output.output_catalog import *


class BackendIPython(BackendBase):
    """
    Common base for the IPython UIs.

    EXAMPLES::

        sage: from sage.repl.rich_output.backend_ipython import BackendIPython
        sage: BackendIPython()._repr_()
        Traceback (most recent call last):
        ...
        NotImplementedError: derived classes must implement this method
    """

    def install(self, **kwds):
        """
        Switch the Sage rich output to the IPython backend.

        INPUT:

        - ``shell`` -- keyword argument; the IPython shell

        No tests since switching away from the doctest rich output
        backend will break the doctests.

        EXAMPLES::

            sage: from sage.repl.interpreter import get_test_shell
            sage: from sage.repl.rich_output.backend_ipython import BackendIPython
            sage: backend = BackendIPython()
            sage: shell = get_test_shell()
            sage: backend.install(shell=shell)
            sage: shell.run_cell('1+1')
            2

        TESTS::

            sage: from sage.repl.rich_output.backend_doctest import BackendDoctest
            sage: from sage.repl.rich_output import get_display_manager
            sage: get_display_manager().switch_backend(BackendDoctest())
            IPython command line
        """
        shell = kwds['shell']
        from sage.repl.display.formatter import SageDisplayFormatter
        shell.display_formatter = SageDisplayFormatter(parent=shell)
        shell.configurables.append(shell.display_formatter)

    def set_underscore_variable(self, obj):
        """
        Set the ``_`` builtin variable.

        Since IPython handles the history itself, this does nothing.

        INPUT:

        - ``obj`` -- anything

        EXAMPLES::

            sage: from sage.repl.interpreter import get_test_shell
            sage: from sage.repl.rich_output.backend_ipython import BackendIPython
            sage: backend = BackendIPython()
            sage: backend.set_underscore_variable(123)
            sage: _
            0
        """
        pass

    def display_immediately(self, plain_text, rich_output):
        """
        Show output immediately.

        This method is similar to the rich output :meth:`displayhook`,
        except that it can be invoked at any time.

        INPUT:

        Same as :meth:`displayhook`.

        OUTPUT: this method does not return anything

        EXAMPLES::

            sage: from sage.repl.rich_output.output_basic import OutputPlainText
            sage: plain_text = OutputPlainText.example()
            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonNotebook
            sage: backend = BackendIPythonNotebook()
            sage: _ = backend.display_immediately(plain_text, plain_text)
            Example plain text output
        """
        formatted, metadata = self.displayhook(plain_text, rich_output)
        if not formatted:
            return
        publish_display_data(data=formatted, metadata=metadata)


class BackendIPythonCommandline(BackendIPython):
    """
    Backend for the IPython Command Line.

    EXAMPLES::

        sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
        sage: BackendIPythonCommandline()
        IPython command line
    """

    def default_preferences(self):
        """
        Return the backend's display preferences.

        The default for the commandline is to not plot graphs since
        the launching of an external viewer is considered too
        disruptive.

        OUTPUT:

        Instance of
        :class:`~sage.repl.rich_output.preferences.DisplayPreferences`.

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: backend.default_preferences()
            Display preferences:
            * align_latex is not specified
            * graphics is not specified
            * supplemental_plot = never
            * text is not specified
        """
        from sage.repl.rich_output.preferences import DisplayPreferences
        return DisplayPreferences(supplemental_plot='never')

    def _repr_(self):
        """
        Return a string representation.

        OUTPUT: string

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: backend._repr_()
            'IPython command line'
        """
        return 'IPython command line'

    def supported_output(self):
        """
        Return the outputs that are supported by the IPython commandline backend.

        OUTPUT:

        Iterable of output container classes, that is, subclass of
        :class:`~sage.repl.rich_output.output_basic.OutputBase`).
        The order is ignored.

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: supp = backend.supported_output();  supp     # random output
            set([<class 'sage.repl.rich_output.output_graphics.OutputImageGif'>,
                 ...,
                 <class 'sage.repl.rich_output.output_graphics.OutputImagePng'>])
            sage: from sage.repl.rich_output.output_basic import OutputLatex
            sage: OutputLatex in supp
            True
        """
        return set([
            OutputPlainText, OutputAsciiArt, OutputUnicodeArt, OutputLatex,
            OutputImagePng, OutputImageGif,
            OutputImagePdf, OutputImageDvi,
            OutputSceneJmol, OutputSceneWavefront, OutputSceneThreejs,
        ])

    def displayhook(self, plain_text, rich_output):
        """
        Backend implementation of the displayhook.

        INPUT:

        - ``plain_text`` -- instance of
          :class:`~sage.repl.rich_output.output_basic.OutputPlainText`. The
          plain text version of the output.

        - ``rich_output`` -- instance of an output container class
          (subclass of
          :class:`~sage.repl.rich_output.output_basic.OutputBase`). Guaranteed
          to be one of the output containers returned from
          :meth:`supported_output`, possibly the same as
          ``plain_text``.

        OUTPUT:

        The IPython commandline display hook returns the IPython
        display data, a pair of dictionaries. The first dictionary
        contains mime types as keys and the respective output as
        value. The second dictionary is metadata.

        EXAMPLES::

            sage: from sage.repl.rich_output.output_basic import OutputPlainText
            sage: plain_text = OutputPlainText.example()
            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: backend.displayhook(plain_text, plain_text)
            ({'text/plain': 'Example plain text output'}, {})
        """
        if isinstance(rich_output, OutputPlainText):
            return ({'text/plain': rich_output.text.get_str()}, {})
        elif isinstance(rich_output, OutputAsciiArt):
            return ({'text/plain': rich_output.ascii_art.get_str()}, {})
        elif isinstance(rich_output, OutputUnicodeArt):
            return ({'text/plain': rich_output.unicode_art.get_str()}, {})
        elif isinstance(rich_output, OutputLatex):
            return ({'text/plain': rich_output.latex.get_str()}, {})
        elif isinstance(rich_output, OutputImagePng):
            msg = self.launch_viewer(
                rich_output.png.filename(ext='png'), plain_text.text.get_str())
            return ({'text/plain': msg}, {})
        elif isinstance(rich_output, OutputImageGif):
            msg = self.launch_viewer(
                rich_output.gif.filename(ext='gif'), plain_text.text.get_str())
            return ({'text/plain': msg}, {})
        elif isinstance(rich_output, OutputImagePdf):
            msg = self.launch_viewer(
                rich_output.pdf.filename(ext='pdf'), plain_text.text.get_str())
            return ({'text/plain': msg}, {})
        elif isinstance(rich_output, OutputImageDvi):
            msg = self.launch_viewer(
                rich_output.dvi.filename(ext='dvi'), plain_text.text.get_str())
            return ({'text/plain': msg}, {})
        elif isinstance(rich_output, OutputSceneJmol):
            msg = self.launch_jmol(rich_output, plain_text.text.get_str())
            return ({'text/plain': msg}, {})
        elif isinstance(rich_output, OutputSceneWavefront):
            msg = self.launch_sage3d(rich_output, plain_text.text.get_str())
            return ({'text/plain': msg}, {})
        elif isinstance(rich_output, OutputSceneThreejs):
            msg = self.launch_viewer(
                rich_output.html.filename(ext='html'), plain_text.text.get_str())
            return ({'text/plain': msg}, {})
        else:
            raise TypeError('rich_output type not supported')

    def display_immediately(self, plain_text, rich_output):
        """
        Show output without going back to the command line prompt.

        This method is similar to the rich output :meth:`displayhook`,
        except that it can be invoked at any time. On the Sage command
        line it launches viewers just like :meth:`displayhook`.

        INPUT:

        Same as :meth:`displayhook`.

        OUTPUT: this method does not return anything

        EXAMPLES::

            sage: from sage.repl.rich_output.output_basic import OutputPlainText
            sage: plain_text = OutputPlainText.example()
            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: backend.display_immediately(plain_text, plain_text)
            Example plain text output
        """
        formatdata, metadata = self.displayhook(plain_text, rich_output)
        print(formatdata['text/plain'])

    def launch_viewer(self, image_file, plain_text):
        """
        Launch external viewer for the graphics file.

        INPUT:

        - ``image_file`` -- string; file name of the image file

        - ``plain_text`` -- string; the plain text representation of
          the image file

        OUTPUT:

        String. Human-readable message indicating whether the viewer
        was launched successfully.

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: backend.launch_viewer('/path/to/foo.bar', 'Graphics object')
            'Launched bar viewer for Graphics object'
        """
        base, dot_ext = os.path.splitext(image_file)
        ext = dot_ext.lstrip(os.path.extsep)
        from sage.misc.viewer import viewer
        command = viewer(ext)
        if not command:
            command = viewer.browser()
        from sage.doctest import DOCTEST_MODE
        if not DOCTEST_MODE:
            os.system('{0} {1} 2>/dev/null 1>/dev/null &'
                      .format(command, image_file))
        return 'Launched {0} viewer for {1}'.format(ext, plain_text)

    def launch_jmol(self, output_jmol, plain_text):
        """
        Launch the stand-alone jmol viewer.

        INPUT:

        - ``output_jmol`` --
          :class:`~sage.repl.rich_output.output_graphics3d.OutputSceneJmol`; the
          scene to launch Jmol with

        - ``plain_text`` -- string; the plain text representation

        OUTPUT:

        string; human-readable message indicating that the viewer was launched.

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: from sage.repl.rich_output.output_graphics3d import OutputSceneJmol
            sage: backend.launch_jmol(OutputSceneJmol.example(), 'Graphics3d object')   # needs sage.plot
            'Launched jmol viewer for Graphics3d object'
        """
        from sage.doctest import DOCTEST_MODE
        from sage.interfaces.jmoldata import JmolData
        jdata = JmolData()
        if not jdata.is_jmol_available() and not DOCTEST_MODE:
            raise RuntimeError('jmol cannot run, no suitable java version found')
        launch_script = output_jmol.launch_script_filename()
        jmol_cmd = 'jmol'
        if not DOCTEST_MODE:
            os.system('{0} {1} 2>/dev/null 1>/dev/null &'
                      .format(jmol_cmd, launch_script))
        return 'Launched jmol viewer for {0}'.format(plain_text)

    def is_in_terminal(self):
        """
        Test whether the UI is meant to run in a terminal.

        See
        :meth:`sage.repl.rich_output.display_manager.DisplayManager.is_in_terminal`
        for details.

        OUTPUT: ``True`` for the IPython commandline

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: backend.is_in_terminal()
            True
        """
        return True

    def threejs_offline_scripts(self):
        """
        Three.js script for the IPython command line.

        OUTPUT: string containing script tag

        EXAMPLES::

            sage: # needs threejs
            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonCommandline
            sage: backend = BackendIPythonCommandline()
            sage: backend.threejs_offline_scripts()
            '...<script ...</script>...'
        """
        from sage.features.threejs import Threejs

        if not Threejs().is_present():
            return ''

        script = Threejs().absolute_filename()

        return '\n<script src="{0}"></script>'.format(script)


IFRAME_TEMPLATE = """
<iframe srcdoc="{escaped_html}"
        width="{width}"
        height="{height}"
        style="border: 0;">
</iframe>
"""


class BackendIPythonNotebook(BackendIPython):
    """
    Backend for the IPython Notebook.

    EXAMPLES::

        sage: from sage.repl.rich_output.backend_ipython import BackendIPythonNotebook
        sage: BackendIPythonNotebook()
        IPython notebook
    """

    def _repr_(self):
        """
        Return string representation of the backend.

        OUTPUT: string

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonNotebook
            sage: backend = BackendIPythonNotebook()
            sage: backend._repr_()
            'IPython notebook'
        """
        return 'IPython notebook'

    def supported_output(self):
        """
        Return the outputs that are supported by the IPython notebook backend.

        OUTPUT:

        Iterable of output container classes, that is, subclass of
        :class:`~sage.repl.rich_output.output_basic.OutputBase`).
        The order is ignored.

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonNotebook
            sage: backend = BackendIPythonNotebook()
            sage: supp = backend.supported_output();  supp     # random output
            set([<class 'sage.repl.rich_output.output_graphics.OutputPlainText'>,
                 ...,
                 <class 'sage.repl.rich_output.output_graphics.OutputImagePdf'>])
            sage: from sage.repl.rich_output.output_basic import OutputLatex
            sage: OutputLatex in supp
            True
            sage: from sage.repl.rich_output.output_graphics import OutputImageGif
            sage: OutputImageGif in supp
            True
        """
        return set([
            OutputPlainText, OutputAsciiArt, OutputUnicodeArt, OutputLatex,
            OutputHtml,
            OutputImagePng, OutputImageGif, OutputImageJpg,
            OutputImageSvg, OutputImagePdf,
            OutputSceneJmol, OutputSceneThreejs,
        ])

    def displayhook(self, plain_text, rich_output):
        """
        Backend implementation of the displayhook.

        INPUT:

        - ``plain_text`` -- instance of
          :class:`~sage.repl.rich_output.output_basic.OutputPlainText`. The
          plain text version of the output.

        - ``rich_output`` -- instance of an output container class
          (subclass of
          :class:`~sage.repl.rich_output.output_basic.OutputBase`). Guaranteed
          to be one of the output containers returned from
          :meth:`supported_output`, possibly the same as
          ``plain_text``.

        OUTPUT:

        The IPython notebook display hook returns the IPython
        display data, a pair of dictionaries. The first dictionary
        contains mime types as keys and the respective output as
        value. The second dictionary is metadata.

        EXAMPLES::

            sage: from sage.repl.rich_output.output_basic import OutputPlainText
            sage: plain_text = OutputPlainText.example()
            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonNotebook
            sage: backend = BackendIPythonNotebook()
            sage: backend.displayhook(plain_text, plain_text)
            ({'text/plain': 'Example plain text output'}, {})
        """
        if isinstance(rich_output, OutputPlainText):
            return ({'text/plain': rich_output.text.get_str()}, {})
        elif isinstance(rich_output, OutputAsciiArt):
            return ({'text/plain': rich_output.ascii_art.get_str()}, {})
        elif isinstance(rich_output, OutputUnicodeArt):
            return ({'text/plain': rich_output.unicode_art.get_str()}, {})
        elif isinstance(rich_output, OutputLatex):
            return ({'text/latex': rich_output.latex.get_str(),
                     'text/plain': plain_text.text.get_str(),
                     }, {})
        elif isinstance(rich_output, OutputHtml):
            data = {'text/html': rich_output.html.get_str(),
                    'text/plain': plain_text.text.get_str()}
            if rich_output.latex:
                data['text/latex'] = rich_output.latex.get_str()
            return (data, {})
        elif isinstance(rich_output, OutputImagePng):
            return ({'image/png': rich_output.png.get(),
                     'text/plain': plain_text.text.get_str(),
                     }, {})
        elif isinstance(rich_output, OutputImageGif):
            return ({'text/html': rich_output.html_fragment(),
                     'text/plain': plain_text.text.get_str(),
                     }, {})
        elif isinstance(rich_output, OutputImageJpg):
            return ({'image/jpeg': rich_output.jpg.get(),
                     'text/plain': plain_text.text.get_str(),
                     }, {})
        elif isinstance(rich_output, OutputImageSvg):
            return ({'image/svg+xml': rich_output.svg.get(),
                     'text/plain': plain_text.text.get_str(),
                     }, {})
        elif isinstance(rich_output, OutputImagePdf):
            return ({'image/png': rich_output.png.get(),
                     'text/plain': plain_text.text.get_str(),
                     }, {})
        elif isinstance(rich_output, OutputSceneJmol):
            from sage.repl.display.jsmol_iframe import JSMolHtml
            jsmol = JSMolHtml(rich_output, height=500)
            return ({'text/html': jsmol.iframe(),
                     'text/plain': plain_text.text.get_str(),
                     }, {})
        elif isinstance(rich_output, OutputSceneThreejs):
            escaped_html = html.escape(rich_output.html.get_str())
            iframe = IFRAME_TEMPLATE.format(
                escaped_html=escaped_html,
                width='100%',
                height=400,
            )
            return ({'text/html': iframe,
                     'text/plain': plain_text.text.get_str(),
                     }, {})
        else:
            raise TypeError('rich_output type not supported')

    def threejs_offline_scripts(self):
        """
        Three.js script for the IPython notebook.

        OUTPUT: string containing script tag

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_ipython import BackendIPythonNotebook
            sage: backend = BackendIPythonNotebook()
            sage: backend.threejs_offline_scripts()                                     # needs sage.plot
            '...<script src="/nbextensions/threejs-sage/r.../three.min.js...<\\/script>...'
        """
        from sage.repl.rich_output import get_display_manager
        from sage.features.threejs import Threejs
        CDN_script = get_display_manager().threejs_scripts(online=True)
        CDN_script = CDN_script.replace('</script>', r'<\/script>').replace('\n', ' \\\n')
        return """
<script src="/nbextensions/threejs-sage/{}/three.min.js"></script>
<script>
  if ( !window.THREE ) document.write('{}');
</script>
        """.format(Threejs().required_version(), CDN_script)



# sage_setup: distribution = sagemath-repl
# sage.doctest: needs sage.plot sage.symbolic
"""
This file (./sol/graphique_doctest.sage) was *autogenerated* from ./sol/graphique.tex,
with sagetex.sty version 2011/05/27 v2.3.1.
It contains the contents of all the sageexample environments from this file.
You should be able to doctest this file with:
sage -t ./sol/graphique_doctest.sage
It is always safe to delete this file; it is not used in typesetting your
document.

Sage example in ./sol/graphique.tex, line 3::

  sage: reset()

Sage example in ./sol/graphique.tex, line 10::

  sage: t = var('t'); liste = [a + cos(t) for a in srange(0, 2, 0.1)]
  sage: g = polar_plot(liste, (t, 0, 2 * pi)); g.show(aspect_ratio = 1)

Sage example in ./sol/graphique.tex, line 38::

  sage: f = lambda x: abs(x**2 - 1/4)
  sage: def liste_pts(u0, n):
  ....:     u = u0; liste = [[u0,0]]
  ....:     for k in range(n):
  ....:         v, u = u, f(u)
  ....:         liste.extend([[v,u], [u,u]])
  ....:     return(liste)
  sage: g = line(liste_pts(1.1, 8), rgbcolor = (.9,0,0))
  sage: g += line(liste_pts(-.4, 8), rgbcolor = (.01,0,0))
  sage: g += line(liste_pts(1.3, 3), rgbcolor = (.5,0,0))
  sage: g += plot(f, -1, 3, rgbcolor = 'blue')
  sage: g += plot(x, -1, 3, rgbcolor = 'green')
  sage: g.show(aspect_ratio = 1, ymin = -.2, ymax = 3)

Sage example in ./sol/graphique.tex, line 78::

  sage: x = var('x'); y = function('y')
  sage: DE = x^2 * diff(y(x), x) - y(x) == 0
  sage: desolve(DE, y(x))
  _C*e^(-1/x)
  sage: g = plot([c*e^(-1/x) for c in srange(-8, 8, 0.4)], (x, -3, 3))
  sage: y = var('y')
  sage: g += plot_vector_field((x^2, y), (x,-3,3), (y,-5,5))
  sage: g.show()  # not tested, known bug, see :issue:`32657`

Sage example in ./sol/graphique.tex, line 124::

  sage: from sage.calculus.desolvers import desolve_system_rk4
  sage: f = lambda x, y: [a*x-b*x*y,-c*y+d*b*x*y]
  sage: x, y, t = var('x, y, t')
  sage: a, b, c, d = 1., 0.1, 1.5, 0.75
  sage: P = desolve_system_rk4(f(x,y), [x,y],\
  ....:        ics=[0,10,5], ivar=t, end_points=15)
  sage: Ql = [[i,j] for i,j,k in P]; p = line(Ql, color='red')
  sage: p += text("Rabbits", (12,37), fontsize=10, color='red')
  sage: Qr = [[i,k] for i,j,k in P]; p += line(Qr, color='blue')
  sage: p += text("Foxes", (12,7), fontsize=10, color='blue')
  sage: p.axes_labels(["time", "population"])
  sage: p.show(gridlines = True)

Sage example in ./sol/graphique.tex, line 154::

  sage: n = 10;  L = srange(6, 18, 12 / n); R = srange(3, 9, 6 / n)
  sage: def g(x, y): v = vector(f(x, y)); return v / v.norm()
  sage: q = plot_vector_field(g(x, y), (x, 0, 60), (y, 0, 36))
  sage: for j in range(n):
  ....:     P = desolve_system_rk4(f(x,y), [x,y],
  ....:            ics=[0,L[j],R[j]], ivar=t, end_points=15)
  ....:     Q = [[j,k] for i,j,k in P]
  ....:     q += line(Q, color=hue(.8-j/(2*n)))
  sage: q.axes_labels(["rabbits", "foxes"]); q.show()

Sage example in ./sol/graphique.tex, line 185::

  sage: from scipy import integrate
  sage: def dX_dt(X, t=0): return [X[1], 0.5*X[1] - X[0] - X[1]^3]
  sage: t = srange(0, 40, 0.01);  x0 = srange(-2, 2, 0.1); y0 = 2.5
  sage: CI = [[i, y0] for i in x0] + [[i, -y0] for i in x0]
  sage: def g(x, y): v = vector(dX_dt([x, y])); return v / v.norm()
  sage: x, y = var('x, y'); n = len(CI)
  sage: q = plot_vector_field(g(x, y), (x, -3, 3), (y, -y0, y0))
  sage: for j in range(n):  # long time
  ....:     X = integrate.odeint(dX_dt, CI[j], t)
  ....:     q += line(X, color=(1.7*j/(4*n),1.5*j/(4*n),1-3*j/(8*n)))
  sage: X = integrate.odeint(dX_dt, [0.01,0], t)
  sage: q += line(X, color = 'red'); q.show()

Sage example in ./sol/graphique.tex, line 234::

  sage: from scipy import integrate
  sage: t = srange(0, 40, 0.2)
  sage: n = 35; CI_cart = [[4, .2 * i] for i in range(n)]
  sage: CI = list(map(lambda x: [sqrt(x[0]^2+x[1]^2),
  ....:      pi - arctan(x[1]/x[0])], CI_cart))
  sage: for alpha in [0.1, 0.5, 1, 1.25]:                      # long time
  ....:     dX_dt = lambda X, t=0: [cos(X[1])*(1-1/X[0]^2),
  ....:             -sin(X[1]) * (1/X[0]+1/X[0]^3) + 2*alpha/X[0]^2]
  ....:     q = circle((0, 0), 1, fill=True, rgbcolor='purple')
  ....:     for j in range(n):
  ....:         X = integrate.odeint(dX_dt, CI[j], t)
  ....:         Y = [[u[0]*cos(u[1]), u[0]*sin(u[1])] for u in X]
  ....:         q += line(Y, xmin = -4, xmax = 4, color='blue')
  ....:     q.show(aspect_ratio = 1, axes = False)
"""



# sage_setup: distribution = sagemath-repl
# sage.doctest: needs sage.rings.finite_rings
##          Sage Doctest File         ##
#**************************************#
#*    Generated from PreTeXt source   *#
#*    on 2017-08-24T11:43:34-07:00    *#
#*                                    *#
#*   http://mathbook.pugetsound.edu   *#
#*                                    *#
#**************************************#
##
"""
Please contact Rob Beezer (beezer@ups.edu) with
any test failures here that need to be changed
as a result of changes accepted into Sage.  You
may edit/change this file in any sensible way, so
that development work may procede.  Your changes
may later be replaced by the authors of "Abstract
Algebra: Theory and Applications" when the text is
updated, and a replacement of this file is proposed
for review.
"""
##
## To execute doctests in these files, run
##   $ $SAGE_ROOT/sage -t <directory-of-these-files>
## or
##   $ $SAGE_ROOT/sage -t <a-single-file>
##
## Replace -t by "-tp n" for parallel testing,
##   "-tp 0" will use a sensible number of threads
##
## See: http://www.sagemath.org/doc/developer/doctesting.html
##   or run  $ $SAGE_ROOT/sage --advanced  for brief help
##
## Generated at 2017-08-24T11:43:34-07:00
## From "Abstract Algebra"
## At commit 26d3cac0b4047f4b8d6f737542be455606e2c4b4
##
## Section 22.6 Sage
##
r"""
~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: F.<a> = GF(7^15); F
    Finite Field in a of size 7^15

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: F.polynomial()
    a^15 + 5*a^6 + 6*a^5 + 6*a^4 + 4*a^3 + a^2 + 2*a + 4

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: a^15 + 5*a^6 + 6*a^5 + 6*a^4 + 4*a^3 + a^2 + 2*a + 4
    0

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: conway_polynomial(7, 15)
    x^15 + 5*x^6 + 6*x^5 + 6*x^4 + 4*x^3 + x^2 + 2*x + 4

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: y = polygen(Integers(7), 'y')
    sage: P = y.parent()
    sage: p = P([4, 5, 2, 6, 3, 3, 6, 2, 1, 1, 2, 5, 6, 3, 5, 1]); p
    y^15 + 5*y^14 + 3*y^13 + 6*y^12 + 5*y^11 + 2*y^10 + y^9 +
    y^8 + 2*y^7 + 6*y^6 + 3*y^5 + 3*y^4 + 6*y^3 + 2*y^2 + 5*y + 4

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: p.is_irreducible()
    True

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: T.<b> = GF(7^15, modulus=p); T
    Finite Field in b of size 7^15

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: F.<a> = GF(5^4)
    sage: a^458
    3*a^3 + 2*a^2 + a + 3

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: (3*a^3 + 2*a^2 + a + 3).log(a)
    458

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: exponent = (3*a^3 + 2*a^2 + a + 3).log(2*a^3 + 4*a^2 + 4*a)
    sage: exponent
    211

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: (2*a^3 + 4*a^2 + 4*a)^exponent == 3*a^3 + 2*a^2 + a + 3
    True

~~~~~~~~~~~~~~~~~~~~~~ ::

    sage: (3*a^3 + 2*a^2 + a + 3).log(a^2 + 4*a + 4)
    Traceback (most recent call last):
    ...
    ValueError: no logarithm exists
"""



# sage_setup: distribution = sagemath-repl
r"""
Display Manager

This is the heart of the rich output system, the display manager
arbitrates between

* Backend capabilities: what can be displayed

* Backend preferences: what gives good quality on the backend

* Sage capabilities: every Sage object can only generate certain
  representations, and

* User preferences: typeset vs. plain text vs. ascii art, etc.

The display manager is a singleton class, Sage always has exactly one
instance of it. Use :func:`get_display_manager` to obtain it.

EXAMPLES::

    sage: from sage.repl.rich_output import get_display_manager
    sage: dm = get_display_manager();  dm
    The Sage display manager using the doctest backend
"""

# ****************************************************************************
#       Copyright (C) 2015 Volker Braun <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import annotations

import warnings
from typing import Any
from typing import Self  # type: ignore

from sage.repl.rich_output.output_basic import (
    OutputAsciiArt,
    OutputPlainText,
    OutputUnicodeArt,
)
from sage.repl.rich_output.output_browser import (
    OutputHtml,
)
from sage.repl.rich_output.preferences import DisplayPreferences
from sage.structure.sage_object import SageObject


class DisplayException(Exception):
    """
    Base exception for all rich output-related exceptions.

    EXAMPLES::

        sage: from sage.repl.rich_output.display_manager import DisplayException
        sage: raise DisplayException('foo')
        Traceback (most recent call last):
        ...
        DisplayException: foo
    """
    pass


class OutputTypeException(DisplayException):
    """
    Wrong Output container.

    The output containers are the subclasses of
    :class:`~sage.repl.rich_output.output_basic.OutputBase` that
    contain the entire output. The display backends must create output
    containers of a suitable type depending on the displayed Python
    object. This exception indicates that there is a mistake in the
    backend and it returned the wrong type of output container.

    EXAMPLES::

        sage: from sage.repl.rich_output.display_manager import OutputTypeException
        sage: raise OutputTypeException('foo')
        Traceback (most recent call last):
        ...
        OutputTypeException: foo
    """
    pass


class RichReprWarning(UserWarning):
    """
    Warning that is throws if a call to ``_rich_repr_`` fails.

    If an object implements ``_rich_repr_`` then it must return a
    value, possibly ``None`` to indicate that no rich output can be
    generated. But it may not raise an exception as it is very
    confusing for the user if the displayhook fails.

    EXAMPLES::

        sage: from sage.repl.rich_output.display_manager import RichReprWarning
        sage: raise RichReprWarning('foo')
        Traceback (most recent call last):
        ...
        RichReprWarning: foo
    """
    pass


class restricted_output:

    def __init__(self, display_manager, output_classes):
        """
        Context manager to temporarily restrict the accepted output types.

        In the context, the output is restricted to the output
        container types listed in ``output_classes``. Additionally,
        display preferences are changed not to show graphics.

        INPUT:

        - ``display_manager`` -- the display manager

        - ``output_classes`` -- iterable of output container types

        EXAMPLES::

            sage: from sage.repl.rich_output.display_manager import (
            ....:     get_display_manager, restricted_output)
            sage: dm = get_display_manager()
            sage: restricted_output(dm, [dm.types.OutputPlainText])
            <sage.repl.rich_output.display_manager.restricted_output object at 0x...>
        """
        self._display_manager = display_manager
        self._output_classes = frozenset(output_classes)

    def __enter__(self):
        """
        Enter the restricted output context.

        EXAMPLES::

            sage: from sage.repl.rich_output.display_manager import (
            ....:     get_display_manager, restricted_output)
            sage: dm = get_display_manager()
            sage: len(dm.supported_output()) > 1
            True
            sage: with restricted_output(dm, [dm.types.OutputPlainText]):
            ....:    dm.supported_output()
            frozenset({<class 'sage.repl.rich_output.output_basic.OutputPlainText'>})

            sage: dm.preferences.supplemental_plot
            'never'
            sage: dm.preferences.supplemental_plot = 'always'
            sage: with restricted_output(dm, [dm.types.OutputPlainText]):
            ....:    dm.preferences
            Display preferences:
            * align_latex is not specified
            * graphics = disable
            * supplemental_plot = never
            * text is not specified
            sage: dm.preferences.supplemental_plot = 'never'
        """
        dm = self._display_manager
        self._original = dm._supported_output
        dm._supported_output = self._output_classes
        self._original_prefs = DisplayPreferences(dm.preferences)
        dm.preferences.graphics = 'disable'
        dm.preferences.supplemental_plot = 'never'

    def __exit__(self, exception_type, value, traceback):
        """
        Exit the restricted output context.

        EXAMPLES::

            sage: from sage.repl.rich_output.display_manager import (
            ....:     get_display_manager, restricted_output)
            sage: dm = get_display_manager()
            sage: with restricted_output(dm, [dm.types.OutputPlainText]):
            ....:     assert len(dm.supported_output()) == 1
            sage: assert len(dm.supported_output()) > 1
        """
        dm = self._display_manager
        dm._supported_output = self._original
        dm.preferences.graphics = self._original_prefs.graphics
        dm.preferences.supplemental_plot = self._original_prefs.supplemental_plot


class DisplayManager(SageObject):

    _instance: Self | None = None

    def __init__(self):
        """
        The Display Manager.

        Used to decide what kind of rich output is best.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: get_display_manager()
            The Sage display manager using the doctest backend
        """
        assert DisplayManager._instance is None
        DisplayManager._instance = self
        from sage.repl.rich_output.backend_base import BackendSimple
        self.switch_backend(BackendSimple())

    @classmethod
    def get_instance(cls) -> Self:
        """
        Get the singleton instance.

        This class method is equivalent to
        :func:`get_display_manager`.

        OUTPUT: the display manager singleton

        EXAMPLES::

            sage: from sage.repl.rich_output.display_manager import DisplayManager
            sage: DisplayManager.get_instance()
            The Sage display manager using the doctest backend
        """
        if cls._instance is not None:
            return cls._instance
        else:
            return cls()

    def _repr_(self):
        """
        Return a string representation.

        OUTPUT: string

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: get_display_manager()
            The Sage display manager using the doctest backend
        """
        s = 'The Sage display manager using the {0} backend'.format(self._backend)
        return s

    @property
    def types(self):
        """
        Catalog of all output container types.

        Note that every output type must be registered in
        :mod:`sage.repl.rich_output.output_catalog`.

        OUTPUT:

        Returns the :mod:`sage.repl.rich_output.output_catalog`
        module.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm.types.OutputPlainText
            <class 'sage.repl.rich_output.output_basic.OutputPlainText'>
        """
        import sage.repl.rich_output.output_catalog
        return sage.repl.rich_output.output_catalog

    def switch_backend(self, backend, **kwds):
        """
        Switch to a new backend.

        INPUT:

        - ``backend`` -- instance of
          :class:`~sage.repl.rich_output.backend_base.BackendBase`

        - ``kwds`` -- optional keyword arguments that are passed on to
          the :meth:`~sage.repl.rich_output.backend_base.BackendBase.install`
          method

        OUTPUT: the previous backend

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_base import BackendSimple
            sage: simple = BackendSimple()
            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager();  dm
            The Sage display manager using the doctest backend

            sage: previous = dm.switch_backend(simple)
            sage: dm
            The Sage display manager using the simple backend

        Restore the doctest backend::

            sage: dm.switch_backend(previous) is simple
            True
        """
        from sage.repl.rich_output.backend_base import BackendBase
        if not isinstance(backend, BackendBase):
            raise ValueError('backend must be instance of BackendBase class')
        supported = backend.supported_output()
        if not any(issubclass(out, OutputPlainText) for out in supported):
            raise ValueError('every backend must support plain text')
        try:
            self._backend.uninstall()
        except AttributeError:
            pass   # first time we switch
        # clear caches
        self._output_promotions = dict()
        self._supported_output = frozenset(
            map(self._demote_output_class, backend.supported_output()))
        # install new backend
        try:
            old_backend = self._backend
        except AttributeError:
            old_backend = None
        self._backend = backend
        self._backend.install(**kwds)
        self._preferences = DisplayPreferences(self._backend.default_preferences())
        return old_backend

    @property
    def preferences(self):
        """
        Return the preferences.

        OUTPUT:

        The display preferences as instance of
        :class:`~sage.repl.rich_output.preferences.DisplayPreferences`.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm.preferences
            Display preferences:
            * align_latex is not specified
            * graphics is not specified
            * supplemental_plot = never
            * text is not specified
        """
        return self._preferences

    def is_in_terminal(self):
        """
        Test whether the UI is meant to run in a terminal.

        When this method returns ``True``, you can assume that it is
        possible to use ``raw_input`` or launch external programs that
        take over the input.

        Otherwise, you should assume that the backend runs remotely or
        in a pty controlled by another program. Then you should not
        launch external programs with a (text or graphical) UI.

        This is used to enable/disable interpreter consoles.

        OUTPUT:

        Boolean.
        """
        return self._backend.is_in_terminal()

    def check_backend_class(self, backend_class):
        """
        Check that the current backend is an instance of
        ``backend_class``.

        This is, for example, used by the Sage IPython display
        formatter to ensure that the IPython backend is in use.

        INPUT:

        - ``backend_class`` -- type of a backend class

        OUTPUT:

        This method returns nothing. A :exc:`RuntimeError` is raised if
        ``backend_class`` is not the type of the current backend.

        EXAMPLES::

            sage: from sage.repl.rich_output.backend_base import BackendSimple
            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm.check_backend_class(BackendSimple)
            Traceback (most recent call last):
            ...
            RuntimeError: check failed: current backend is invalid
        """
        if not isinstance(self._backend, backend_class):
            raise RuntimeError('check failed: current backend is invalid')

    def _demote_output_class(self, output_class):
        """
        Helper for :meth:`switch_backend`.

        INPUT:

        - ``output_class`` -- a possibly derived class from one of the
          output container classes in :meth:`types`

        OUTPUT: the underlying container class that it was derived from

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm._demote_output_class(dm.types.OutputPlainText)
            <class 'sage.repl.rich_output.output_basic.OutputPlainText'>
        """
        from sage.repl.rich_output.output_basic import OutputBase
        if not issubclass(output_class, OutputBase):
            raise OutputTypeException(
                'invalid output container type: {0} is not subclass of OutputBase'
                .format(output_class))
        result = None
        for type_name in dir(self.types):
            if type_name.startswith('_'):
                continue
            tp = getattr(self.types, type_name)
            if not issubclass(tp, OutputBase):
                continue
            if issubclass(output_class, tp):
                if result is not None:
                    raise OutputTypeException(
                        '{0} inherits from multiple output classes'
                        .format(output_class))
                else:
                    self._output_promotions[tp] = output_class
                    result = tp
        if result is None:
            raise OutputTypeException(
                '{0} does not inherit from any known output class'
                .format(output_class))
        return result

    def _promote_output(self, output):
        """
        Promote output container to a backend-specific subclass.

        INPUT:

        - ``output`` -- instance of a subclass of
          :class:`~sage.repl.rich_output.output_basic.OutputBase`. Backend-agnostic
          output container.

        OUTPUT:

        If the backend returned a subclass of the type of ``output``:
        promote the class to the subclass and return it. Otherwise,
        the unchanged output is returned.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: out = dm._promote_output(dm.types.OutputPlainText('test'))
            sage: type(out)
            <class 'sage.repl.rich_output.output_basic.OutputPlainText'>
        """
        try:
            specialized_class = self._output_promotions[type(output)]
        except KeyError:
            return output
        output.__class__ = specialized_class
        return output

    def _preferred_text_formatter(self, obj, plain_text=None, **kwds):
        r"""
        Return the preferred textual representation.

        INPUT:

        - ``obj`` -- anything; the objects to format

        - ``plain_text`` -- ``None`` (default) or string. The plain
          text representation. If specified, this will be used for
          plain text output.

        OUTPUT:

        One of
        :class:`~sage.repl.rich_output.output_basic.OutputPlainText`,
        :class:`~sage.repl.rich_output.output_basic.OutputAsciiArt`,
        :class:`~sage.repl.rich_output.output_basic.OutputUnicodeArt`,
        or
        :class:`~sage.repl.rich_output.output_basic.OutputLatex`
        containing the preferred textual representation of ``obj``
        supported by the backend.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm.preferences.text is None
            True
            sage: dm._preferred_text_formatter([1/42])
            OutputPlainText container

            sage: dm.preferences.text = 'plain'
            sage: dm._preferred_text_formatter([1/42])
            OutputPlainText container

            sage: dm.preferences.text = 'ascii_art'
            sage: dm._preferred_text_formatter([1/42])
            OutputAsciiArt container

            sage: dm.preferences.text = 'unicode_art'
            sage: dm._preferred_text_formatter([1/42])
            OutputUnicodeArt container

            sage: dm.preferences.text = 'latex'
            sage: dm._preferred_text_formatter([1/42])  # doctest backend does not support html
            OutputPlainText container

            sage: del dm.preferences.text   # reset to default
        """
        want = self.preferences.text
        supported = self._backend.supported_output()
        if want == 'ascii_art' and OutputAsciiArt in supported:
            out = self._backend.ascii_art_formatter(obj, **kwds)
            if type(out) is not OutputAsciiArt:
                raise OutputTypeException('backend returned wrong output type, require AsciiArt')
            return out
        if want == 'unicode_art' and OutputUnicodeArt in supported:
            out = self._backend.unicode_art_formatter(obj, **kwds)
            if type(out) is not OutputUnicodeArt:
                raise OutputTypeException('backend returned wrong output type, require UnicodeArt')
            return out
        if want == 'latex' and OutputHtml in supported:
            out = self._backend.latex_formatter(obj, **kwds)
            if type(out) is not OutputHtml:
                raise OutputTypeException('backend returned wrong output type, require Html')
            return out
        if plain_text is not None:
            if type(plain_text) is not OutputPlainText:
                raise OutputTypeException('backend returned wrong output type, require PlainText')
            return plain_text
        out = self._backend.plain_text_formatter(obj, **kwds)
        if type(out) is not OutputPlainText:
            raise OutputTypeException('backend returned wrong output type, require PlainText')
        return out

    def _call_rich_repr(self, obj, rich_repr_kwds):
        """
        Call the ``_rich_repr_`` method.

        This method calls ``obj._rich_repr_``. If this raises an
        exception, it is caught and a suitable warning is displayed.

        INPUT:

        - ``obj`` -- anything

        - ``rich_repr_kwds`` -- dictionary; optional keyword arguments
          that are passed through to ``obj._rich_repr_``

        OUTPUT:

        Whatever ``_rich_repr_`` returned. If it raises an exception,
        then a :class:`DisplayFormatterWarning` is displayed and
        ``None`` is returned.

        EXAMPLES::

            sage: class Foo(SageObject):
            ....:     def _rich_repr_(self, display_manager, **kwds):
            ....:         raise ValueError('reason')

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm._call_rich_repr(Foo(), {})
            doctest:...: ...RichReprWarning: Exception in _rich_repr_ while displaying object: reason
        """
        if rich_repr_kwds:
            # do not ignore errors from invalid options
            return obj._rich_repr_(self, **rich_repr_kwds)
        try:
            return obj._rich_repr_(self)
        except NotImplementedError:
            # don't warn on NotImplementedErrors
            return None
        except Exception as e:
            warnings.warn(
                'Exception in _rich_repr_ while displaying object: {0}'.format(e),
                RichReprWarning,
            )

    def _rich_output_formatter(self, obj, rich_repr_kwds):
        """
        Generate appropriate rich output containers.

        INPUT:

        - ``obj`` -- anything

        - ``rich_repr_kwds`` -- dictionary; optional keyword arguments
          that are passed through to ``obj._rich_repr_``

        OUTPUT:

        A pair of rich output containers. The first is plain text,
        that is, an instance of
        :class:`~sage.repl.rich_output.output_basic.OutputPlainText`. The
        second is the best rich output available, subject to the
        constraints of the backend.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm._rich_output_formatter(1/2, dict())
            (OutputPlainText container, OutputPlainText container)
        """
        rich_output = None
        plain_text = None
        has_rich_repr = isinstance(obj, SageObject) and hasattr(obj, '_rich_repr_')
        if has_rich_repr:
            rich_output = self._call_rich_repr(obj, rich_repr_kwds)
        if isinstance(rich_output, OutputPlainText):
            plain_text = rich_output
        elif has_rich_repr:
            with restricted_output(self, [OutputPlainText]):
                plain_text = self._call_rich_repr(obj, rich_repr_kwds)
        if plain_text is None:
            plain_text = self._backend.plain_text_formatter(obj, **rich_repr_kwds)
        if rich_output is None:
            rich_output = self._preferred_text_formatter(
                obj, plain_text=plain_text, **rich_repr_kwds)
        # promote output container types to backend-specific containers
        plain_text = self._promote_output(plain_text)
        rich_output = self._promote_output(rich_output)
        # check that the output container types are valid for the backend
        supported = self._backend.supported_output()
        if type(plain_text) not in supported:
            raise OutputTypeException(
                'text output container not supported: {0}'.format(type(plain_text)))
        if type(rich_output) not in supported:
            raise OutputTypeException(
                'output container not supported: {0}'.format(type(rich_output)))
        return plain_text, rich_output

    def graphics_from_save(self, save_function, save_kwds,
                           file_extension, output_container,
                           figsize=None, dpi=None):
        r"""
        Helper to construct graphics.

        This method can be used to simplify the implementation of a
        ``_rich_repr_`` method of a graphics object if there is
        already a function to save graphics to a file.

        INPUT:

        - ``save_function`` -- callable that can save graphics to a file
          and accepts options like :meth:`sage.plot.graphics.Graphics.save`

        - ``save_kwds`` -- dictionary; keyword arguments that are
          passed to the save function

        - ``file_extension`` -- string starting with ``'.'``; the file
          extension of the graphics file

        - ``output_container`` -- subclass of
          :class:`sage.repl.rich_output.output_basic.OutputBase`. The
          output container to use. Must be one of the types in
          :meth:`supported_output`.

        - ``figsize`` -- pair of integers (optional). The desired graphics
          size in pixels. Suggested, but need not be respected by the
          output.

        - ``dpi`` -- integer (optional). The desired resolution in dots
          per inch. Suggested, but need not be respected by the output.

        OUTPUT: an instance of ``output_container``

        EXAMPLES::

            sage: # needs sage.plot sage.symbolic
            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: plt = plot(sin)
            sage: out = dm.graphics_from_save(plt.save, dict(), '.png',
            ....:                             dm.types.OutputImagePng)
            sage: out
            OutputImagePng container
            sage: out.png.get().startswith(b'\x89PNG')
            True
            sage: out.png.filename()   # random
            '/home/user/.sage/temp/localhost.localdomain/23903/tmp_pu5woK.png'
        """
        import os
        if not file_extension.startswith(os.path.extsep):
            raise ValueError('file_extension must start with a period')
        if output_container not in self.supported_output():
            raise OutputTypeException('output_container is not supported by backend')
        from sage.misc.temporary_file import tmp_filename
        filename = tmp_filename(ext=file_extension)
        # Call the save_function with the right arguments
        kwds = dict(save_kwds)
        if figsize is not None:
            kwds['figsize'] = figsize
        if dpi is not None:
            kwds['dpi'] = dpi
        save_function(filename, **kwds)
        from sage.repl.rich_output.buffer import OutputBuffer
        buf = OutputBuffer.from_file(filename)
        return output_container(buf)

    def threejs_scripts(self, online):
        """
        Return Three.js script tag for the current backend.

        INPUT:

        - ``online`` -- boolean determining script usage context

        OUTPUT: string containing script tag

        .. NOTE::

            This base method handles ``online=True`` case only, serving CDN
            script tag. Location of script for offline usage is
            backend-specific.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: get_display_manager().threejs_scripts(online=True)                    # needs sage.plot
            '...<script src="https://cdn.jsdelivr.net/gh/sagemath/threejs-sage@...'
            sage: get_display_manager().threejs_scripts(online=False)                   # needs sage.plot
            Traceback (most recent call last):
            ...
            ValueError: current backend does not support
            offline threejs graphics
        """
        from sage.features.threejs import Threejs
        if online:
            version = Threejs().required_version()
            return """
<script src="https://cdn.jsdelivr.net/gh/sagemath/threejs-sage@{0}/build/three.min.js"></script>
            """.format(version)
        try:
            return self._backend.threejs_offline_scripts()
        except AttributeError:
            raise ValueError(
                'current backend does not support offline threejs graphics')

    def supported_output(self):
        """
        Return the output container classes that can be used.

        OUTPUT:

        Frozen set of subclasses of
        :class:`~sage.repl.rich_output.output_basic.OutputBase`. If
        the backend defines derived container classes, this method
        will always return their base classes.

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm.types.OutputPlainText in dm.supported_output()
            True
            sage: type(dm.supported_output())
            <... 'frozenset'>
        """
        return self._supported_output

    def displayhook(self, obj: Any) -> Any | None:
        """
        Implementation of the displayhook.

        Every backend must pass the value of the last statement of a
        line / cell to this method. See also
        :meth:`display_immediately` if you want do display rich output
        while a program is running.

        INPUT:

        - ``obj`` -- anything; the object to be shown

        OUTPUT: whatever the backend's displayhook method returned

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm.displayhook(1/2)
            1/2
        """
        if obj is None:
            return
        self._backend.set_underscore_variable(obj)
        plain_text, rich_output = self._rich_output_formatter(obj, dict())
        return self._backend.displayhook(plain_text, rich_output)

    def display_immediately(self, obj, **rich_repr_kwds):
        """
        Show output without going back to the command line prompt.

        This method must be called to create rich output from an
        object when we are not returning to the command line prompt,
        for example during program execution. Typically, it is being
        called by :meth:`sage.plot.graphics.Graphics.show`.

        INPUT:

        - ``obj`` -- anything; the object to be shown

        - ``rich_repr_kwds`` -- optional keyword arguments that are
          passed through to ``obj._rich_repr_``

        EXAMPLES::

            sage: from sage.repl.rich_output import get_display_manager
            sage: dm = get_display_manager()
            sage: dm.display_immediately(1/2)
            1/2
        """
        plain_text, rich_output = self._rich_output_formatter(obj, rich_repr_kwds)
        self._backend.display_immediately(plain_text, rich_output)


get_display_manager = DisplayManager.get_instance



# sage_setup: distribution = sagemath-repl
"""
Detecting external software

This module makes up a list of external software that Sage interfaces. Availability
of each software is tested only when necessary. This is mainly used for the doctests
which require certain external software installed on the system.

Even though the functions in this module should also work when an external
software is not present, most doctests in this module are only tested if
testing of external software is explicitly enabled in order to avoid invoking
external software otherwise. See :issue:`28819` for details.

AUTHORS:

- Kwankyu Lee (2016-03-09) -- initial version, based on code by Robert Bradshaw and Nathann Cohen
"""

# ****************************************************************************
#       Copyright (C) 2016      Kwankyu Lee <ekwankyu@gmail.com>
#                     2018      Thierry Monteil
#                     2018-2021 Sébastien Labbé
#                     2019      Markus Wageringel
#                     2020      John H. Palmieri
#                     2021      Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sys


# With OS X, Python 3.8 defaults to use 'spawn' instead of 'fork' in
# multiprocessing, and Sage doctesting doesn't work with 'spawn'. See
# trac #27754.
# With Python 3.14, the default changed to 'forkserver' on Linux as well.
# Sage doctesting requires 'fork' method.
if sys.platform not in ('win32', 'emscripten'):
    import multiprocessing
    multiprocessing.set_start_method('fork', force=True)

# Functions in this module whose name is of the form 'has_xxx' tests if the
# software xxx is available to Sage.
prefix = 'has_'


def has_internet() -> bool:
    """
    Test if Internet is available.

    Failure of connecting to the site "https://www.sagemath.org" within a second
    is regarded as internet being not available.

    EXAMPLES::

        sage: from sage.doctest.external import has_internet
        sage: has_internet()  # random, optional -- internet
        FeatureTestResult('internet', True)
    """
    from sage.features.internet import Internet
    return Internet().is_present()


def has_latex() -> bool:
    """
    Test if Latex is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_latex
        sage: has_latex() # optional - latex
        FeatureTestResult('latex', True)
    """
    from sage.features.latex import latex
    return latex().is_present()


def has_xelatex() -> bool:
    """
    Test if xelatex is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_xelatex
        sage: has_xelatex()   # optional - xelatex
        FeatureTestResult('xelatex', True)
    """
    from sage.features.latex import xelatex
    return xelatex().is_present()


def has_pdflatex() -> bool:
    """
    Test if pdflatex is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_pdflatex
        sage: has_pdflatex()   # optional - pdflatex
        FeatureTestResult('pdflatex', True)
    """
    from sage.features.latex import pdflatex
    return pdflatex().is_present()


def has_lualatex() -> bool:
    """
    Test if lualatex is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_lualatex
        sage: has_lualatex()   # optional - lualatex
        FeatureTestResult('lualatex', True)
    """
    from sage.features.latex import lualatex
    return lualatex().is_present()


def has_magma() -> bool:
    """
    Test if Magma is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_magma
        sage: has_magma() # random, optional - magma
        True
    """
    from sage.features.interfaces import Magma
    return Magma().is_present()


def has_matlab() -> bool:
    """
    Test if Matlab is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_matlab
        sage: has_matlab() # random, optional - matlab
        True
    """
    from sage.features.interfaces import Matlab
    return Matlab().is_present()


def has_mathematica() -> bool:
    """
    Test if Mathematica is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_mathematica
        sage: has_mathematica() # random, optional - mathematica
        True
    """
    from sage.features.interfaces import Mathematica
    return Mathematica().is_present()


def has_maple() -> bool:
    """
    Test if Maple is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_maple
        sage: has_maple() # random, optional - maple
        True
    """
    from sage.features.interfaces import Maple
    return Maple().is_present()


def has_macaulay2() -> bool:
    """
    Test if Macaulay2 is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_macaulay2
        sage: has_macaulay2() # random, optional - macaulay2
        True
    """
    from sage.features.interfaces import Macaulay2
    return Macaulay2().is_present()


def has_octave() -> bool:
    """
    Test if Octave is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_octave
        sage: has_octave() # random, optional - octave
        True
    """
    from sage.features.interfaces import Octave
    return Octave().is_present()


def has_pandoc() -> bool:
    """
    Test if pandoc is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_pandoc
        sage: has_pandoc()      # optional -- pandoc
        FeatureTestResult('pandoc', True)
    """
    from sage.features.pandoc import Pandoc
    return Pandoc().is_present()


def has_scilab() -> bool:
    """
    Test if Scilab is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_scilab
        sage: has_scilab() # random, optional - scilab
        True
    """
    from sage.interfaces.scilab import scilab
    try:
        scilab('2+3')
        return True
    except Exception:
        return False


def has_cplex() -> bool:
    """
    Test if CPLEX is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_cplex
        sage: has_cplex() # random, optional - CPLEX
        FeatureTestResult('cplex', True)
    """
    from sage.features.mip_backends import CPLEX
    return CPLEX().is_present()


def has_gurobi() -> bool:
    """
    Test if Gurobi is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_gurobi
        sage: has_gurobi() # random, optional - Gurobi
        FeatureTestResult('gurobi', True)
    """
    from sage.features.mip_backends import Gurobi
    return Gurobi().is_present()


def has_graphviz() -> bool:
    """
    Test if graphviz (dot, twopi, neato) are available.

    EXAMPLES::

        sage: from sage.doctest.external import has_graphviz
        sage: has_graphviz()   # optional -- graphviz
        FeatureTestResult('graphviz', True)
    """
    from sage.features.graphviz import Graphviz
    return Graphviz().is_present()


def has_ffmpeg() -> bool:
    """
    Test if ffmpeg is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_ffmpeg
        sage: has_ffmpeg()      # optional -- ffmpeg
        FeatureTestResult('ffmpeg', True)
    """
    from sage.features.ffmpeg import FFmpeg
    return FFmpeg().is_present()


def has_imagemagick() -> bool:
    """
    Test if ImageMagick (command magick or convert) is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_imagemagick
        sage: has_imagemagick() # optional -- imagemagick
        FeatureTestResult('imagemagick', True)
    """
    from sage.features.imagemagick import ImageMagick
    return ImageMagick().is_present()


def has_dvipng() -> bool:
    """
    Test if dvipng is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_dvipng
        sage: has_dvipng() # optional -- dvipng
        FeatureTestResult('dvipng', True)
    """
    from sage.features.dvipng import dvipng
    return dvipng().is_present()


def has_pdf2svg() -> bool:
    """
    Test if pdf2svg is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_pdf2svg
        sage: has_pdf2svg() # optional -- pdf2svg
        FeatureTestResult('pdf2svg', True)
    """
    from sage.features.pdf2svg import pdf2svg
    return pdf2svg().is_present()


def has_rubiks() -> bool:
    """
    Test if the rubiks package (``cu2``, ``cubex``, ``dikcube``,
    ``mcube``, ``optimal``, and ``size222``) is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_rubiks
        sage: has_rubiks()   # optional -- rubiks
        FeatureTestResult('rubiks', True)
    """
    from sage.features.rubiks import Rubiks
    return Rubiks().is_present()


def has_4ti2() -> bool:
    """
    Test if the 4ti2 package is available.

    EXAMPLES::

        sage: from sage.doctest.external import has_4ti2
        sage: has_4ti2()   # optional -- 4ti2
        FeatureTestResult('4ti2', True)
    """
    from sage.features.four_ti_2 import FourTi2
    return FourTi2().is_present()


def external_features():
    r"""
    Generate the features that are only to be tested if ``--optional=external`` is used.

    .. SEEALSO::

        :func:`sage.features.all.all_features`

    EXAMPLES::

        sage: from sage.doctest.external import external_features
        sage: next(external_features())
        Feature('internet')
    """
    from sage.features.internet import Internet
    yield Internet()
    import sage.features.latex
    yield from sage.features.latex.all_features()
    import sage.features.ffmpeg
    yield from sage.features.ffmpeg.all_features()
    import sage.features.interfaces
    for feature in sage.features.interfaces.all_features():
        if feature.name not in ('mathics', 'regina'):
            yield feature
    from sage.features.mip_backends import CPLEX, Gurobi
    yield CPLEX()
    yield Gurobi()


def _external_software() -> list[str]:
    """
    Return the alphabetical list of external software supported by this module.

    EXAMPLES::

        sage: from sage.doctest.external import external_software
        sage: sorted(external_software) == external_software
        True
    """
    return sorted(f.name for f in external_features())


external_software: list[str] = _external_software()


class AvailableSoftware:
    """
    This class keeps the set of available software whose availability is detected lazily
    from the list of external software.

    EXAMPLES::

        sage: from sage.doctest.external import external_software, available_software
        sage: external_software
        ['cplex',
         'dvips',
         'ffmpeg',
         'gurobi',
         'internet',
         'latex',
         'latex_package_tkz_graph',
         'lualatex',
         'macaulay2',
         'magma',
         'maple',
         'mathematica',
         'matlab',
         'octave',
         'pdflatex',
         'scilab',
         'xelatex']
        sage: 'internet' in available_software # random, optional - internet
        True
        sage: available_software.issuperset(set(['internet','latex'])) # random, optional - internet latex
        True
    """
    def __init__(self):
        """
        Initialization.

        EXAMPLES::

            sage: from sage.doctest.external import AvailableSoftware
            sage: S = AvailableSoftware()
            sage: S.seen() # random
            []
        """
        self._allow_external = True
        # For multiprocessing of doctests, the data self._seen should be
        # shared among subprocesses. Thus we use Array class from the
        # multiprocessing module.
        from sage.features.all import all_features
        self._external_features = set(external_features())
        features = set(self._external_features)
        features.update(all_features())
        self._features = sorted(features, key=lambda feature: feature.name)
        self._indices = {feature.name: idx for idx, feature in enumerate(self._features)}
        try:
            from multiprocessing import Array
            self._seen = Array('i', len(self._features)) # initialized to zeroes
            self._hidden = Array('i', len(self._features)) # initialized to zeroes
        except ImportError:  # module '_multiprocessing' is removed in Pyodide due to browser limitations
            self._seen = [0] * len(self._features)
            self._hidden = [0] * len(self._features)

    def __contains__(self, item):
        """
        Return ``True`` if ``item`` is available on the system.

        EXAMPLES::

            sage: from sage.doctest.external import available_software
            sage: 'internet' in available_software # random, optional - internet
            True
        """
        try:
            idx = self._indices[item]
        except KeyError:
            return False
        feature = self._features[idx]
        if feature.is_hidden():
            if not self._hidden[idx]:
                self._hidden[idx] = 1
            available = False  # a hidden feature is considered to be not available
        else:
            if not self._allow_external and feature in self._external_features:
                # an external feature is considered to be not available
                # if this is not allowed
                available = False
            elif feature.is_present():
                available = True
            else:
                available = False
        if available:
            if not self._seen[idx]:
                self._seen[idx] = 1
            return True
        else:
            return False

    def issuperset(self, other):
        """
        Return ``True`` if ``other`` is a subset of ``self``.

        EXAMPLES::

            sage: from sage.doctest.external import available_software
            sage: available_software.issuperset(set(['internet','latex','magma'])) # random, optional - internet latex magma
            True
        """
        return all(item in self for item in other)

    def detectable(self):
        """
        Return the list of names of those features for which testing their presence is allowed.
        """
        return [feature.name
                for feature, seen in zip(self._features, self._seen)
                if seen >= 0 and (self._allow_external or feature not in self._external_features)]

    def seen(self):
        """
        Return the list of detected external software.

        EXAMPLES::

            sage: from sage.doctest.external import available_software
            sage: available_software.seen() # random
            ['internet', 'latex', 'magma']
        """
        return [feature.name
                for feature, seen in zip(self._features, self._seen)
                if seen > 0]

    def hidden(self):
        """
        Return the list of detected hidden external software.

        EXAMPLES::

            sage: # needs conway_polynomials database_cremona_mini_ellcurve database_ellcurves database_graphs
            sage: from sage.doctest.external import available_software
            sage: from sage.features.databases import all_features
            sage: for f in all_features():
            ....:    f.hide()
            ....:    if f._spkg_type() == 'standard':
            ....:         test = f.name in available_software
            ....:    f.unhide()
            sage: sorted(available_software.hidden())
            [...'conway_polynomials',...
             'database_cremona_mini_ellcurve',...
             'database_ellcurves',...
             'database_graphs'...]
        """
        return [feature.name
                for feature, hidden in zip(self._features, self._hidden)
                if hidden > 0]


available_software = AvailableSoftware()



# sage_setup: distribution = sagemath-repl
"""
All imports for Jupyter
"""

from sage.all_cmdline import *

from sage.repl.ipython_kernel.widgets_sagenb import (input_box, text_control, slider,
                                                     range_slider, checkbox, selector, input_grid, color_selector)
from sage.repl.ipython_kernel.interact import interact

from pathlib import Path


# sage_setup: distribution = sagemath-repl
# sage.doctest: needs sage.symbolic
"""
This file (./sol/polynomes_doctest.sage) was *autogenerated* from ./sol/polynomes.tex,
with sagetex.sty version 2011/05/27 v2.3.1.
It contains the contents of all the sageexample environments from this file.
You should be able to doctest this file with:
sage -t ./sol/polynomes_doctest.sage
It is always safe to delete this file; it is not used in typesetting your
document.

Sage example in ./sol/polynomes.tex, line 10::

  sage: x = polygen(QQ, 'y'); y = polygen(QQ, 'x')

Sage example in ./sol/polynomes.tex, line 78::

  sage: T = sage.symbolic.function_factory.function('T', nargs=2)
  sage: def to_chebyshev_basis(pol):
  ....:     (x,) = pol.variables()
  ....:     res = 0
  ....:     for n in range(pol.degree(), -1, -1):
  ....:         quo, pol = pol.quo_rem(chebyshev_T(n, x))
  ....:         res += quo * T(n, x)
  ....:     return res

Sage example in ./sol/polynomes.tex, line 105::

  sage: x = polygen(QQ)
  sage: p = 4*x^6 + 4*x^5 + 1/9*x^4 - 2*x^3 + 2/19*x^2 + 1
  sage: p_cheb = to_chebyshev_basis(p); p_cheb
  1/8*T(6, x) + 1/4*T(5, x) + 55/72*T(4, x) + 3/4*T(3, x) +
  2713/1368*T(2, x) + T(1, x) + 1069/456*T(0, x)
  sage: p_cheb.substitute_function(T, chebyshev_T).expand()
  4*x^6 + 4*x^5 + 1/9*x^4 - 2*x^3 + 2/19*x^2 + 1

Sage example in ./sol/polynomes.tex, line 131::

  sage: def mydiv(u, v, n):
  ....:     v0 = v.constant_coefficient()
  ....:     quo = 0; rem = u
  ....:     for k in range(n+1):
  ....:         c = rem[0]/v0
  ....:         rem = (rem - c*v) >> 1  # shifting the coefficients
  ....:         quo += c*x^k
  ....:     return quo, rem

Sage example in ./sol/polynomes.tex, line 161::

  sage: def mydiv2(u, v, n):
  ....:     x = u.parent().gen()
  ....:     quo = (u / (v + O(x^(n+1)))).polynomial()
  ....:     rem = (u - quo*v) >> (n+1)
  ....:     return quo, rem

Sage example in ./sol/polynomes.tex, line 179::

  sage: x = polygen(QQ)
  sage: u = -5*x^10 + 7/2*x^9 + 3*x^8 - x^6 - 1/3*x^5 - x^4 - 1/2*x^3 + 15*x^2 - 1/42*x + 2
  sage: v = 3*x^4 + 3/8*x^3 - 1/2*x^2 + x + 5/2
  sage: n = 8
  sage: q, r = mydiv(u, v, n)
  sage: u - q*v - x^(n+1)*r
  0
  sage: q, r = mydiv2(u, v, n)
  sage: u - q*v - x^(n+1)*r
  0

Sage example in ./sol/polynomes.tex, line 296::

  sage: Poly.<x> = Integers(10^5)[]
  sage: P = x^1000 - 23*x^729 + 5*x^2 - 12*x - 7
  sage: Quo.<s> = Poly.quo(P)
  sage: op = s^(10^10000)  # long time
  sage: add(op[n]*(n+7) for n in range(1000))  # long time
  63477

Sage example in ./sol/polynomes.tex, line 383::

  sage: from sage.matrix.berlekamp_massey import berlekamp_massey
  sage: berlekamp_massey([1, 1, 2, 3, 8, 11, 34, 39, 148, 127])
  x^3 - 5*x + 2

Sage example in ./sol/polynomes.tex, line 408::

  sage: R.<x> = GF(17)[]
  sage: pairs = [(0,-1), (1,0), (2,7), (3,5)]
  sage: s = R(QQ['x'].lagrange_polynomial(pairs)); s
  6*x^3 + 2*x^2 + 10*x + 16
  sage: [s(i) for i in range(4)]
  [16, 0, 7, 5]

Sage example in ./sol/polynomes.tex, line 428::

  sage: s.rational_reconstruction(mul(x-i for i in range(4)), 1, 2)
  (15*x + 2, x^2 + 11*x + 15)

Sage example in ./sol/polynomes.tex, line 454::

  sage: S.<x> = PowerSeriesRing(QQ)
  sage: t = S(0)
  sage: for i in range(7): # here t is correct up to degree 2i+1
  ....:     # with O(x^15) we prevent the truncation order to grow
  ....:     t = (1+t^2).integral() + O(x^15)
  sage: t
  x + 1/3*x^3 + 2/15*x^5 + 17/315*x^7 + 62/2835*x^9 + 1382/155925*x^11
  + 21844/6081075*x^13 + O(x^15)
"""
